# Short Time Fourier Transform
# Copyright (C) 2019  Eric Bezzam, Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

"""Class for real-time STFT analysis and processing."""
from __future__ import division

import numpy as np
from numpy.lib.stride_tricks import as_strided as _as_strided
from .dft import DFT


class STFT(object):
    """
    A class for STFT processing.

    Parameters
    -----------
    N : int
        number of samples per frame
    hop : int
        hop size
    analysis_window : numpy array
        window applied to block before analysis
    synthesis_window : numpy array
        window applied to the block before synthesis
    channels : int
        number of signals

    transform : str, optional
        which FFT package to use: 'numpy' (default), 'pyfftw', or 'mkl'
    streaming : bool, optional
        whether (True, default) or not (False) to "stitch" samples between
        repeated calls of 'analysis' and 'synthesis' if we are receiving a 
        continuous stream of samples.
    num_frames : int, optional
        Number of frames to be processed. If set, this will be strictly enforced
        as the STFT block will allocate memory accordingly. If not set, there
        will be no check on the number of frames sent to 
        analysis/process/synthesis

        NOTE:
            1) num_frames = 0, corresponds to a "real-time" case in which each
            input block corresponds to [hop] samples.
            2) num_frames > 0, requires [(num_frames-1)*hop + N] samples as the
            last frame must contain [N] samples.

    precision : string, np.float32, np.float64, np.complex64, np.complex128, optional
        How many precision bits to use for the input.
        If 'single'/np.float32/np.complex64, 32 bits for real inputs or 64 for complex spectrum.
        Otherwise, cast to 64 bits for real inputs or 128 for complex spectrum (default).

    """

    def __init__(self, N, hop=None, analysis_window=None,
                 synthesis_window=None, channels=1, transform='numpy',
                 streaming=True, precision='double', **kwargs):

        # initialize parameters
        self.num_samples = N            # number of samples per frame
        self.num_channels = channels    # number of channels
        self.mono = True if self.num_channels == 1 else False
        if hop is not None:             # hop size --> number of input samples
            self.hop = hop  
        else:
            self.hop = self.num_samples

        if precision == np.float32 or precision == np.complex64 or precision == 'single':
            self.time_dtype = np.float32
            self.freq_dtype = np.complex64
        else:
            self.time_dtype = np.float64
            self.freq_dtype = np.complex128

        # analysis and synthesis window
        self.analysis_window = analysis_window
        self.synthesis_window = synthesis_window

        # prepare variables for DFT object
        self.transform = transform
        self.nfft = self.num_samples   # differ when there is zero-padding
        self.nbin = self.nfft // 2 + 1

        # initialize filter + zero padding --> use set_filter
        self.zf = 0
        self.zb = 0
        self.H = None           # filter frequency spectrum
        self.H_multi = None     # for multiple frames

        # check keywords
        if 'num_frames' in kwargs.keys():
            self.fixed_input = True
            num_frames = kwargs['num_frames']
            if num_frames < 0:
                raise ValueError('num_frames must be non-negative!')
            self.num_frames = num_frames
        else:
            self.fixed_input = False
            self.num_frames = 0

        # allocate all the required buffers
        self.streaming = streaming
        self._make_buffers()

    def _make_buffers(self):
        """
        Allocate memory for internal buffers according to FFT size, number of
        channels, and number of frames.
        """

        # state variables
        self.n_state = self.num_samples - self.hop
        self.n_state_out = self.nfft - self.hop

        # make DFT object
        self.dft = DFT(nfft=self.nfft,
                       D=self.num_channels,
                       analysis_window=self.analysis_window,
                       synthesis_window=self.synthesis_window,
                       transform=self.transform)
        """
        1D array for num_channels=1 as the FFTW package can only take 1D array 
        for 1D DFT.
        """
        if self.mono:
            # input buffer
            self.fft_in_buffer = np.zeros(self.nfft, dtype=self.time_dtype)
            # state buffer
            self.x_p = np.zeros(self.n_state, dtype=self.time_dtype)
            # prev reconstructed samples
            self.y_p = np.zeros(self.n_state_out, dtype=self.time_dtype)
            # output samples
            self.out = np.zeros(self.hop, dtype=self.time_dtype)
        else:
            # input buffer
            self.fft_in_buffer = np.zeros((self.nfft, self.num_channels),
                                          dtype=self.time_dtype)
            # state buffer
            self.x_p = np.zeros((self.n_state, self.num_channels),
                                dtype=self.time_dtype)
            # prev reconstructed samples
            self.y_p = np.zeros((self.n_state_out, self.num_channels),
                                dtype=self.time_dtype)
            # output samples
            self.out = np.zeros((self.hop, self.num_channels),
                                dtype=self.time_dtype)

        # useful views on the input buffer
        self.fft_in_state = self.fft_in_buffer[self.zf:self.zf + self.n_state, ]
        self.fresh_samples = self.fft_in_buffer[self.zf + self.n_state:
                                                self.zf + self.n_state +
                                                self.hop, ]
        self.old_samples = self.fft_in_buffer[self.zf + self.hop:
                                              self.zf + self.hop +
                                              self.n_state, ]

        # if fixed number of frames to process
        if self.fixed_input:
            if self.num_frames == 0:
                if self.mono:
                    self.X = np.zeros(self.nbin, dtype=self.freq_dtype)
                else:
                    self.X = np.zeros((self.nbin, self.num_channels),
                                      dtype=self.freq_dtype)
            else:
                self.X = np.squeeze(np.zeros((self.num_frames,
                                              self.nbin,
                                              self.num_channels),
                                             dtype=self.freq_dtype))
                # DFT object for multiple frames
                self.dft_frames = DFT(nfft=self.nfft, D=self.num_frames,
                                      analysis_window=self.analysis_window,
                                      synthesis_window=self.synthesis_window,
                                      transform=self.transform)
        else: # we will allocate these on-the-fly
            self.X = None
            self.dft_frames = None

    def reset(self):
        """
        Reset state variables. Necessary after changing or setting the filter
        or zero padding.
        """

        if self.mono:
            self.fft_in_buffer[:] = 0.
            self.x_p[:] = 0.
            self.y_p[:] = 0.
            self.X[:] = 0.
            self.out[:] = 0.
        else:
            self.fft_in_buffer[:, :] = 0.
            self.x_p[:, :] = 0.
            self.y_p[:, :] = 0.
            self.X[:, :] = 0.
            self.out[:, :] = 0.

    def zero_pad_front(self, zf):
        """
        Set zero-padding at beginning of frame.
        """
        self.zf = zf
        self.nfft = self.num_samples+self.zb+self.zf
        self.nbin = self.nfft//2+1
        if self.analysis_window is not None:
            self.analysis_window = np.concatenate((np.zeros(zf),
                                                   self.analysis_window))
        if self.synthesis_window is not None:
            self.synthesis_window = np.concatenate((np.zeros(zf),
                                                    self.synthesis_window))

        # We need to reallocate buffers after changing zero padding
        self._make_buffers()

    def zero_pad_back(self, zb):
        """
        Set zero-padding at end of frame.
        """
        self.zb = zb
        self.nfft = self.num_samples+self.zb+self.zf
        self.nbin = self.nfft//2+1
        if self.analysis_window is not None:
            self.analysis_window = np.concatenate((self.analysis_window,
                                                   np.zeros(zb)))
        if self.synthesis_window is not None:
            self.synthesis_window = np.concatenate((self.synthesis_window,
                                                    np.zeros(zb)))

        # We need to reallocate buffers after changing zero padding
        self._make_buffers()

    def set_filter(self, coeff, zb=None, zf=None, freq=False):
        """
        Set time-domain FIR filter with appropriate zero-padding.
        Frequency spectrum of the filter is computed and set for the object.
        There is also a check for sufficient zero-padding.

        Parameters
        -----------
        coeff : numpy array
            Filter in time domain.
        zb : int
            Amount of zero-padding added to back/end of frame.
        zf : int
            Amount of zero-padding added to front/beginning of frame.
        freq : bool
            Whether or not given coefficients (coeff) are in the frequency
            domain.
        """
        # apply zero-padding
        if zb is not None:
            self.zero_pad_back(zb)
        if zf is not None:
            self.zero_pad_front(zf)
        if not freq:
            # compute filter magnitude and phase spectrum
            self.H = self.freq_dtype(np.fft.rfft(coeff, self.nfft, axis=0))

            # check for sufficient zero-padding
            if self.nfft < (self.num_samples+len(coeff)-1):
                raise ValueError('Insufficient zero-padding for chosen number '
                                 'of samples per frame (L) and filter length '
                                 '(h). Require zero-padding such that new '
                                 'length is at least (L+h-1).')
        else:
            if len(coeff) != self.nbin:
                raise ValueError('Invalid length for frequency domain '
                                 'coefficients.')
            self.H = coeff

        # prepare filter if fixed input case
        if self.fixed_input:
            if self.num_channels == 1:
                self.H_multi = np.tile(self.H, (self.num_frames, 1))
            else:
                self.H_multi = np.tile(self.H, (self.num_frames, 1, 1))

    def analysis(self, x):
        """
        Parameters
        -----------
        x  : 2D numpy array, [samples, channels]
            Time-domain signal.
        """

        # ----check correct number of channels
        x_shape = x.shape
        if not self.mono:
            if len(x_shape) < 1:   # received mono
                raise ValueError('Received 1-channel signal. Expecting %d '
                                 'channels.' % self.num_channels)
            if x_shape[1] != self.num_channels:
                raise ValueError('Incorrect number of channels. Received %d, '
                                 'expecting %d.' % (x_shape[1],
                                                    self.num_channels))
        else:   # expecting mono
            if len(x_shape) > 1:    # received multi-channel
                raise ValueError('Received %d channels; expecting 1D mono '
                                 'signal.' % x_shape[1])

        # ----check number of frames
        if self.streaming:  # need integer multiple of hops

            if self.fixed_input:
                if x_shape[0] != self.num_frames*self.hop:
                    raise ValueError('Input must be of length %d; received %d '
                                     'samples.' % (self.num_frames*self.hop,
                                                   x_shape[0]))
            else:
                self.num_frames = int(np.ceil(x_shape[0]/self.hop))
                extra_samples = (self.num_frames*self.hop)-x_shape[0]
                if extra_samples:
                    if self.mono:
                        x = np.concatenate((x, np.zeros(extra_samples)))
                    else:
                        x = np.concatenate(
                            (x, np.zeros((extra_samples, self.num_channels))))

        # non-streaming
        # need at least num_samples for last frame
        # e.g.[hop|hop|...|hop|num_samples]
        else:

            if self.fixed_input:
                if x_shape[0] != (self.hop*(self.num_frames-1)+
                                  self.num_samples):
                    raise ValueError('Input must be of length %d; received %d '
                                     'samples.' % ((self.hop*(self.num_frames-1)
                                                    + self.num_samples),
                                                   x_shape[0]))
            else:
                if x_shape[0] < self.num_samples:
                    # raise ValueError('Not enough samples. Received %d; need \
                    #     at least %d.' % (x_shape[0],self.num_samples))
                    extra_samples = self.num_samples - x_shape[0]
                    if self.mono:
                        x = np.concatenate((x, np.zeros(extra_samples)))
                    else:
                        x = np.concatenate(
                            (x, np.zeros((extra_samples, self.num_channels))))
                    self.num_frames = 1
                else:

                    # calculate num_frames and append zeros if necessary
                    self.num_frames = \
                        int(np.ceil((x_shape[0]-self.num_samples)/self.hop) + 1)
                    extra_samples = ((self.num_frames-1)*self.hop +
                                     self.num_samples)-x_shape[0]
                    if extra_samples:
                        if self.mono:
                            x = np.concatenate((x, np.zeros(extra_samples)))
                        else:
                            x = np.concatenate(
                                (x,
                                 np.zeros((extra_samples, self.num_channels))))

        # ----allocate memory if necessary
        if not self.fixed_input:
            self.X = np.squeeze(np.zeros((self.num_frames,
                                          self.nbin,
                                          self.num_channels),
                                         dtype=self.freq_dtype))
            self.dft_frames = DFT(nfft=self.nfft,
                                  D=self.num_frames,
                                  analysis_window=self.analysis_window,
                                  synthesis_window=self.synthesis_window,
                                  transform=self.transform)

        # ----use appropriate function
        if self.streaming:
            self._analysis_streaming(x)
        else:
            self.reset()
            self._analysis_non_streaming(x)
                
        return self.X

    def _analysis_single(self, x_n):
        """
        Transform new samples to STFT domain for analysis.

        Parameters
        -----------
        x_n : numpy array
            [self.hop] new samples
        """

        # correct input size check in: dft.analysis()
        self.fresh_samples[:, ] = x_n[:, ]  # introduce new samples
        self.x_p[:, ] = self.old_samples   # save next state

        # apply DFT to current frame
        self.X[:] = self.dft.analysis(self.fft_in_buffer)

        # shift backwards in the buffer the state
        self.fft_in_state[:, ] = self.x_p[:, ]

    def _analysis_streaming(self, x):
        """
        STFT analysis for streaming case in which we expect
        [num_frames*hop] samples
        """

        if self.num_frames == 1:
            self._analysis_single(x)
        else:
            n = 0
            for k in range(self.num_frames):
                # introduce new samples
                self.fresh_samples[:, ] = x[n:n+self.hop, ]
                # save next state
                self.x_p[:, ] = self.old_samples

                # apply DFT to current frame
                self.X[k, ] = self.dft.analysis(self.fft_in_buffer)

                # shift backwards in the buffer the state
                self.fft_in_state[:, ] = self.x_p[:, ]

                n += self.hop

    def _analysis_non_streaming(self, x):
        """
        STFT analysis for non-streaming case in which we expect
        [(num_frames-1)*hop+num_samples] samples
        """

        ## ----- STRIDED WAY
        new_strides = (x.strides[0], self.hop * x.strides[0])
        new_shape = (self.num_samples, self.num_frames)

        if not self.mono:
            for c in range(self.num_channels):

                y = _as_strided(x[:, c], shape=new_shape, strides=new_strides)
                y = np.concatenate((np.zeros((self.zf, self.num_frames)), y,
                                    np.zeros((self.zb, self.num_frames))))

                if self.num_frames == 1:
                    self.X[:, c] = self.dft_frames.analysis(y[:, 0]).T
                else:
                    self.X[:, :, c] = self.dft_frames.analysis(y).T
        else:

            y = _as_strided(x, shape=new_shape, strides=new_strides)
            y = np.concatenate((np.zeros((self.zf, self.num_frames)), y,
                                np.zeros((self.zb, self.num_frames))))

            if self.num_frames == 1:
                self.X[:] = self.dft_frames.analysis(y[:, 0]).T
            else:
                self.X[:] = self.dft_frames.analysis(y).T

    def _check_input_frequency_dimensions(self, X):
        """
        Ensure that given frequency data is valid, i.e. number of channels and
        number of frequency bins.

        If fixed_input=True, ensure expected number of frames. Otherwise, infer 
        from given data.

        Axis order of X should be : [frames, frequencies, channels]
        """

        # check number of frames and correct number of bins
        X_shape = X.shape
        if len(X_shape) == 1:  # single channel, one frame
            num_frames = 1
        elif len(X_shape) == 2 and not self.mono: # multi-channel, one frame
            num_frames = 1
        elif len(X_shape) == 2 and self.mono: # single channel, multiple frames
            num_frames = X_shape[0]
        elif len(X_shape) == 3 and not self.mono: # multi-channel, multiple frames
            num_frames = X_shape[0]
        else:
            raise ValueError('Invalid input shape.')

        # check number of bins
        if num_frames == 1:
            if X_shape[0] != self.nbin:
                raise ValueError('Invalid number of frequency bins! Expecting '
                                 '%d, got %d.' % (self.nbin, X_shape[0]))
        else:
            if X_shape[1] != self.nbin:
                raise ValueError('Invalid number of frequency bins! Expecting'
                                 ' %d, got %d.' % (self.nbin, X_shape[1]))

        # check number of frames, if fixed input size
        if self.fixed_input:
            if num_frames != self.num_frames:
                raise ValueError('Input must have %d frames!' % self.num_frames)
            self.X[:] = X  # reset if size is alright
        else:
            self.X = X
            self.num_frames = num_frames

        return self.X

    def process(self, X=None):

        """
        Parameters
        -----------
        X  : numpy array
            X can take on multiple shapes:
            1) (N,) if it is single channel and only one frame
            2) (N,D) if it is multi-channel and only one frame
            3) (F,N) if it is single channel but multiple frames
            4) (F,N,D) if it is multi-channel and multiple frames

        Returns
        -----------
        x_r : numpy array
            Reconstructed time-domain signal.

        """

        # check that there is filter
        if self.H is None:
            return

        if X is not None:
            self._check_input_frequency_dimensions(X)

        # use appropriate function
        if self.num_frames == 1:
            self._process_single()
        elif self.num_frames > 1:
            self._process_multiple()

        return self.X

    def _process_single(self):

        np.multiply(self.X, self.H, self.X)

    def _process_multiple(self):

        if not self.fixed_input:
            if self.mono:
                self.H_multi = np.tile(self.H, (self.num_frames, 1))
            else:
                self.H_multi = np.tile(self.H, (self.num_frames, 1, 1))

        np.multiply(self.X, self.H_multi, self.X)

    def synthesis(self, X=None):
        """
        Parameters
        -----------
        X  : numpy array of frequency content
            X can take on multiple shapes:
            1) (N,) if it is single channel and only one frame
            2) (N,D) if it is multi-channel and only one frame
            3) (F,N) if it is single channel but multiple frames
            4) (F,N,D) if it is multi-channel and multiple frames
            where:
            - F is the number of frames
            - N is the number of frequency bins
            - D is the number of channels


        Returns
        -----------
        x_r : numpy array
            Reconstructed time-domain signal.

        """

        if X is not None:
            self._check_input_frequency_dimensions(X)

        # use appropriate function
        if self.num_frames == 1:
            return self._synthesis_single()
        elif self.num_frames > 1:
            return self._synthesis_multiple()

    def _synthesis_single(self):
        """
        Transform to time domain and reconstruct output with overlap-and-add.

        Returns
        -------
        numpy array
            Reconstructed array of samples of length <self.hop>.
        """

        # apply IDFT to current frame
        self.dft.synthesis(self.X)

        return self._overlap_and_add()

    def _synthesis_multiple(self):
        """
        Apply STFT analysis to multiple frames.

        Returns
        -----------
        x_r : numpy array
            Recovered signal.

        """

        # synthesis + overlap and add
        if not self.mono:

            x_r = np.zeros((self.num_frames*self.hop, self.num_channels),
                           dtype=self.time_dtype)

            n = 0
            for f in range(self.num_frames):

                # apply IDFT to current frame and reconstruct output
                x_r[n:n+self.hop, ] = self._overlap_and_add(
                    self.dft.synthesis(self.X[f, :, :]))
                n += self.hop

        else:

            x_r = np.zeros(self.num_frames*self.hop, dtype=self.time_dtype)

            # treat number of frames as the multiple channels for DFT
            if not self.fixed_input:
                self.dft_frames = DFT(nfft=self.nfft,
                                      D=self.num_frames,
                                      analysis_window=self.analysis_window,
                                      synthesis_window=self.synthesis_window,
                                      transform=self.transform)

            # back to time domain
            mx = self.dft_frames.synthesis(self.X.T)

            # overlap and add
            n = 0
            for f in range(self.num_frames):
                x_r[n:n+self.hop, ] = self._overlap_and_add(mx[:, f])
                n += self.hop

        return x_r

    def _overlap_and_add(self, x=None):

        if x is None:
            x = self.dft.x

        self.out[:, ] = x[0:self.hop, ]  # fresh output samples

        # add state from previous frames when overlap is used
        if self.n_state_out > 0:
            m = np.minimum(self.hop, self.n_state_out)
            self.out[:m, ] += self.y_p[:m, ]
            # update state variables
            self.y_p[:-self.hop, ] = self.y_p[self.hop:, ]  # shift out left
            self.y_p[-self.hop:, ] = 0.
            self.y_p[:, ] += x[-self.n_state_out:, ]

        return self.out


" ---------------------------------------------------------------------------- "
" --------------- One-shot functions to avoid creating object. --------------- "
" ---------------------------------------------------------------------------- "
# Authors: Robin Scheibler, Ivan Dokmanic, Sidney Barthe

def analysis(x, L, hop, win=None, zp_back=0, zp_front=0):
    """
    Convenience function for one-shot STFT

    Parameters
    ----------
    x: array_like, (n_samples) or (n_samples, n_channels)
        input signal 
    L: int
        frame size
    hop: int
        shift size between frames
    win: array_like
        the window to apply (default None)
    zp_back: int
        zero padding to apply at the end of the frame
    zp_front: int
        zero padding to apply at the beginning of the frame

    Returns
    -------
    X: ndarray, (n_frames, n_frequencies) or (n_frames, n_frequencies, n_channels)
        The STFT of x
    """

    if x.ndim == 2:
        channels = x.shape[1]
    else:
        channels = 1

    the_stft = STFT(L, hop=hop, analysis_window=win, channels=channels, precision=x.dtype)

    if zp_back > 0:
        the_stft.zero_pad_back(zp_back)

    if zp_front > 0:
        the_stft.zero_pad_front(zp_front)

    # apply transform
    return the_stft.analysis(x)


# inverse STFT
def synthesis(X, L, hop, win=None, zp_back=0, zp_front=0):
    """
    Convenience function for one-shot inverse STFT

    Parameters
    ----------
    X: array_like (n_frames, n_frequencies) or (n_frames, n_frequencies, n_channels)
        The data
    L: int
        frame size
    hop: int
        shift size between frames
    win: array_like
        the window to apply (default None)
    zp_back: int
        zero padding to apply at the end of the frame
    zp_front: int
        zero padding to apply at the beginning of the frame

    Returns
    -------
    x: ndarray, (n_samples) or (n_samples, n_channels)
        The inverse STFT of X
    """

    if X.ndim == 3:
        channels = X.shape[2]
    else:
        channels = 1

    the_stft = STFT(L, hop=hop, synthesis_window=win, channels=channels, precision=X.dtype)

    if zp_back > 0:
        the_stft.zero_pad_back(zp_back)

    if zp_front > 0:
        the_stft.zero_pad_front(zp_front)

    # apply transform
    return the_stft.synthesis(X)


def compute_synthesis_window(analysis_window, hop):
    """
    Computes the optimal synthesis window given an analysis window
    and hop (frame shift). The procedure is described in

    D. Griffin and J. Lim, *Signal estimation from modified short-time Fourier transform,*
    IEEE Trans. Acoustics, Speech, and Signal Process.,
    vol. 32, no. 2, pp. 236-243, 1984.

    Parameters
    ----------
    analysis_window: array_like
        The analysis window
    hop: int
        The frame shift
    """

    norm = np.zeros_like(analysis_window)
    L = analysis_window.shape[0]

    # move the window back as far as possible while still overlapping
    n = 0
    while n - hop > -L:
        n -= hop

    # now move the window and sum all the contributions
    while n < L:
        if n == 0:
            norm += analysis_window ** 2
        elif n < 0:
            norm[:n+L] += analysis_window[-n-L:] ** 2
        else:
            norm[n:] += analysis_window[:-n] ** 2

        n += hop

    return analysis_window / norm




def calculate_speed_of_sound(t, h, p): 
    ''' 
    Compute the speed of sound as a function of
    temperature, humidity and pressure

    Arguments
    ---------

    t: temperature [Celsius]
    h: relative humidity [%]
    p: atmospheric pressure [kpa]

    Return
    ------

    Speed of sound in [m/s]
    '''

    # using crude approximation for now
    return 331.4 + 0.6*t + 0.0124*h




import numpy as np
from scipy.stats import norm as _norm
import pyroomacoustics as pra

def test_median():

    # simple tests
    x = np.arange(1, 11)
    m = pra.median(x)
    assert m == 5.5

    x = np.arange(1, 12)
    m = pra.median(x)
    assert m == 6

    # test dimensions
    x = np.random.rand(10,9,8)

    m, ci = pra.median(x, alpha=0.05)
    assert m.shape == (10,9)
    assert ci.shape == (2,10,9)

    m, ci = pra.median(x, alpha=0.05, axis=1)
    assert m.shape == (10,8)
    assert ci.shape == (2,10,8)

    m, ci = pra.median(x, alpha=0.05, axis=1, keepdims=True)
    assert m.shape == (10,1,8)
    assert ci.shape == (2,10,1,8)


    # Now test statistical property of the confidence interval
    np.random.seed(0)  # fix this for repeatable outcome
    R = 200
    N = [10, 100, 1000]
    alpha = [0.05, 0.01]

    from scipy.stats import uniform
    dist = uniform()
    true_median = dist.median()

    for n in N:
        for a in alpha:

            failures = np.zeros(R, dtype=bool)

            for r in range(R):
                x = dist.rvs(size=n)
                m,ci = pra.median(x, alpha=a)

                if true_median < m+ci[0] or m+ci[1] < true_median:
                    failures[r] = 1

            assert sum(failures) / R <= 1.2 * a  # allow some slack for realizations


if __name__ == '__main__':

    test_median()





from __future__ import division, print_function

import numpy as np
import pyroomacoustics as pra
from pyroomacoustics.transform import STFT
import time
import warnings

# test signal
np.random.seed(0)
num_mic = 25
signals = np.random.randn(100000, num_mic).astype(np.float32)
fs = 16000

# STFT parameters
block_size = 512
hop = block_size//2
win = pra.hann(block_size)
x_r = np.zeros(signals.shape)
num_times = 50

print()

"""
One frame at a time
"""
print("Averaging computation time over %d cases of %d channels of %d samples (%0.1f s at %0.1f kHz)." 
    % (num_times,num_mic,len(signals),(len(signals)/fs),fs/1000) )
print()
print("----- SINGLE FRAME AT A TIME -----")
print("With STFT object (not fixed) : ", end="")
stft = STFT(block_size, hop=hop, channels=num_mic,
    streaming=True, analysis_window=win)
start = time.time()
for k in range(num_times):

    x_r = np.zeros(signals.shape)
    n = 0
    while  signals.shape[0] - n > hop:
        stft.analysis(signals[n:n+hop,])
        x_r[n:n+hop,] = stft.synthesis()
        n += hop
avg_time = (time.time()-start)/num_times
print("%0.3f sec" % avg_time)
err_dB = 20*np.log10(np.max(np.abs(signals[:n-hop,] - x_r[hop:n,])))
print("Error [dB] : %0.3f" % err_dB)


print("With STFT object (fixed) : ", end="")
stft = STFT(block_size, hop=hop, channels=num_mic, num_frames=1, 
    streaming=True, analysis_window=win)
start = time.time()
for k in range(num_times):

    x_r = np.zeros(signals.shape)
    n = 0
    while  signals.shape[0] - n > hop:
        stft.analysis(signals[n:n+hop,])
        x_r[n:n+hop,] = stft.synthesis()
        n += hop
avg_time = (time.time()-start)/num_times
print("%0.3f sec" % avg_time)
err_dB = 20*np.log10(np.max(np.abs(signals[:n-hop,] - x_r[hop:n,])))
print("Error [dB] : %0.3f" % err_dB)


"""
Multiple frame at a time (non-streaming)
"""
print()
print("----- MULTIPLE FRAMES AT A TIME -----")

warnings.filterwarnings("ignore") # to avoid warning of appending zeros to be printed
print("With STFT object (not fixed) : ", end="")
stft = STFT(block_size, hop=hop, channels=num_mic,
    analysis_window=win, streaming=False)
start = time.time()
for k in range(num_times):

    stft.analysis(signals)
    x_r = stft.synthesis()

avg_time = (time.time()-start)/num_times
print("%0.3f sec" % avg_time)
err_dB = 20*np.log10(np.max(np.abs(signals[hop:len(x_r)-hop] - 
    x_r[hop:len(x_r)-hop])))
print("Error [dB] : %0.3f" % err_dB)
warnings.filterwarnings("default")


print("With STFT object (fixed) : ", end="")
num_frames = (len(signals)-block_size)//hop + 1
stft = STFT(block_size, hop=hop, channels=num_mic,
    num_frames=num_frames, analysis_window=win, streaming=False)
start = time.time()
for k in range(num_times):

    stft.analysis(signals[:(num_frames-1)*hop+block_size,:])
    x_r = stft.synthesis()
avg_time = (time.time()-start)/num_times
print("%0.3f sec" % avg_time)
err_dB = 20*np.log10(np.max(np.abs(signals[hop:len(x_r)-hop] - 
    x_r[hop:len(x_r)-hop])))
print("Error [dB] : %0.3f" % err_dB)





# Test of Room constructor
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
import pyroomacoustics as pra

wall_corners = [
    np.array([[0, 3, 3, 0], [0, 0, 0, 0], [0, 0, 2, 2]]),  # left
    np.array([[0, 0, 6, 6], [8, 8, 8, 8], [0, 4, 4, 0]]),  # right
    np.array([[0, 0, 6, 3], [0, 8, 8, 0], [0, 0, 0, 0]]),  # floor
    np.array([[0, 3, 6, 0], [0, 0, 8, 8], [2, 2, 4, 4]]),  # ceiling
    np.array([[0, 0, 0, 0], [0, 0, 8, 8], [0, 2, 4, 0]]),  # back
    np.array([[3, 6, 6, 3], [0, 8, 8, 0], [0, 0, 4, 2]]),  # front
]

absorptions = [0.1, 0.25, 0.25, 0.25, 0.2, 0.15]
scatterings = [0.1, 0.05, 0.1, 0.02, 0.2, 0.3]


def test_room_construct():

    walls = [
        pra.wall_factory(c, [a], [s])
        for c, a, s in zip(wall_corners, absorptions, scatterings)
    ]
    obstructing_walls = []
    microphones = np.array([[1], [2], [1]])

    room = pra.libroom.Room(
        walls,
        [],
        [],
        pra.constants.get("c"),  # speed of sound
        3,
        1e-7,  # energy_thres
        1.0,  # time_thres
        0.5,  # receiver_radius
        0.004,  # hist_bin_size
        True,  # a priori we will always use a hybrid model
    )

    print(room.max_dist)

    return room


if __name__ == "__main__":

    room = test_room_construct()



# Single Channel Noise Removal using Spectral Subtraction
# Copyright (C) 2019  Eric Bezzam, Alexis Mermet
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.


import numpy as np


class SpectralSub(object):
    """
    Here we have a class for performing **single channel** noise reduction via
    spectral subtraction. The instantaneous signal energy and noise floor is
    estimated at each time instance (for each frequency bin) and this is used
    to compute a gain filter with which to perform spectral subtraction.

    For a given frame `n`, the gain for frequency bin `k` is given by:

    .. math::

        G[k, n] = \max \\left \{ \\left ( \dfrac{P[k, n]-\\beta P_N[k, n]}{P[k, n]} \\right )^\\alpha, G_{min} \\right \},

    where :math:`G_{min} = 10^{-(db\_reduc/20)}` and :math:`db\_reduc` is the
    maximum reduction (in dB) that we are willing to perform for each bin (a
    high value can actually be detrimental, see below). The instantaneous
    energy :math:`P[k,n]` is computed by simply squaring the frequency
    amplitude at the bin `k`. The time-frequency decomposition of the input
    signal is typically done with the STFT and overlapping frames. The noise
    estimate :math:`P_N[k, n]` for frequency bin `k` is given by looking back a
    certain number of frames :math:`L` and selecting the bin with the lowest
    energy:

    .. math::

        P_N[k, n] = \min_{[n-L, n]} P[k, n]

    This approach works best when the SNR is positive and the noise is rather
    stationary. An alternative approach for the noise estimate (also in the
    case of stationary noise) would be to apply a lowpass filter for each
    frequency bin.

    With a large suppression, i.e. large values for :math:`db\_reduc`, we can
    observe a typical artefact of such spectral subtraction approaches, namely
    "musical noise".
    `Here <https://www.vocal.com/noise-reduction/musical-noise/>`_ is nice
    article about noise reduction and musical noise.

    Adjusting the constants :math:`\\beta` and :math:`\\alpha` also presents a
    trade-off between suppression and undesirable artefacts, i.e. more
    noticeable musical noise.

    Below is an example of how to use this class to emulate a streaming/online
    input. A full example can be found
    `here <https://github.com/LCAV/pyroomacoustics/blob/master/examples/noise_reduction_spectral_subtraction.py>`__.

    ::

        # initialize STFT and SpectralSub objects
        nfft = 512
        stft = pra.transform.STFT(nfft, hop=nfft//2,
                                  analysis_window=pra.hann(nfft))
        scnr = pra.denoise.SpectralSub(nfft, db_reduc=10, lookback=5,
                                       beta=20, alpha=3)

        # apply block-by-block
        for n in range(num_blocks):

            # go to frequency domain for noise reduction
            stft.analysis(mono_noisy)
            gain_filt = scnr.compute_gain_filter(stft.X)

            # estimating input convolved with unknown response
            mono_denoised = stft.synthesis(gain_filt*stft.X)


    There also exists a "one-shot" function.

    ::

        # import or create `noisy_signal`
        denoised_signal = apply_spectral_sub(noisy_signal, nfft=512,
                                             db_reduc=10, lookback=5,
                                             beta=20, alpha=3)


    Parameters
    ----------
    nfft: int
        FFT size. Length of gain filter, i.e. the number of frequency bins, is
        given by ``nfft//2+1``.
    db_reduc: float
        Maximum reduction in dB for each bin.
    lookback: int
        How many frames to look back for the noise estimate.
    beta: float
        Overestimation factor to "push" the gain filter value (at each
        frequency) closer to the dB reduction specified by ``db_reduc``.
    alpha: float, optional
        Exponent factor to modify transition behavior towards the dB reduction 
        specified by ``db_reduc``. Default is 1.

    """

    def __init__(self, nfft, db_reduc, lookback, beta, alpha=1):

        self.beta = beta
        self.alpha = alpha

        self.n_bins = nfft//2+1
        self.p_prev = np.zeros((self.n_bins, lookback+1))
        self.gmin = 10**(-db_reduc/20)

        self.p_sn = np.zeros(self.n_bins)
        self.p_n = np.zeros(self.n_bins)

    def compute_gain_filter(self, X):
        """
        Parameters
        ----------
        X: numpy array
            Complex spectrum of length ``nfft//2+1``.

        Returns
        -------
        numpy array
            Gain filter to multiply given spectrum with.
        """

        # estimate of signal + noise at current time
        self.p_sn[:] = np.real(np.conj(X)*X)

        # estimate of noise level
        self.p_prev[:, -1] = self.p_sn
        self.p_n[:] = np.min(self.p_prev, axis=1)

        # compute gain filter
        gain_filter = [max((max(self.p_sn[k]-self.beta*self.p_n[k], 0) /
                            self.p_sn[k])**self.alpha, self.gmin)
                       for k in range(self.n_bins)]

        # update
        self.p_prev = np.roll(self.p_prev, -1, axis=1)

        return gain_filter


def apply_spectral_sub(noisy_signal, nfft=512, db_reduc=25, lookback=12,
                       beta=30, alpha=1):
    """
    One-shot function to apply spectral subtraction approach.

    Parameters
    ----------
    noisy_signal : numpy array
        Real signal in time domain.
    nfft: int
        FFT size. Length of gain filter, i.e. the number of frequency bins, is
        given by ``nfft//2+1``.
    db_reduc: float
        Maximum reduction in dB for each bin.
    lookback: int
        How many frames to look back for the noise estimate.
    beta: float
        Overestimation factor to "push" the gain filter value (at each
        frequency) closer to the dB reduction specified by ``db_reduc``.
    alpha: float, optional
        Exponent factor to modify transition behavior towards the dB reduction
        specified by ``db_reduc``. Default is 1.

    Returns
    -------
    numpy array
        Enhanced/denoised signal.
    """

    from pyroomacoustics import hann
    from pyroomacoustics.transform import STFT

    hop = nfft // 2
    window = hann(nfft, flag='asymmetric', length='full')
    stft = STFT(nfft, hop=hop, analysis_window=window, streaming=True)
    scnr = SpectralSub(nfft, db_reduc, lookback, beta, alpha)

    processed_audio = np.zeros(noisy_signal.shape)
    n = 0
    while noisy_signal.shape[0] - n >= hop:
        # SCNR in frequency domain
        stft.analysis(noisy_signal[n:(n + hop), ])
        gain_filt = scnr.compute_gain_filter(stft.X)

        # back to time domain
        processed_audio[n:n + hop, ] = stft.synthesis(gain_filt * stft.X)

        # update step
        n += hop

    return processed_audio




from unittest import TestCase
import numpy as np
from scipy.signal import fftconvolve

import pyroomacoustics as pra

# fix seed for repeatability
np.random.seed(0)

h_len = 30
x_len = 1000
SNR = 1000.  # decibels

h_lp = np.fft.irfft(np.ones(5), n=h_len)
h_rand = np.random.randn(h_len)
h_hann = pra.hann(h_len, flag='symmetric')

x = np.random.randn(x_len)
noise = np.random.randn(x_len + h_len - 1)

def generate_signals(SNR, x, h, noise):
    ''' run convolution '''

    # noise standard deviation
    sigma_noise = 10**(-SNR / 20.)

    y = fftconvolve(x, h) 
    y += sigma_noise * noise

    return y, sigma_noise

class TestDeconvolution(TestCase):

    def test_deconvolve_hann_noiseless(self):

        h = h_hann
        h_len = h_hann.shape[0]
        SNR = 1000.
        tol = 1e-7

        y, sigma_noise = generate_signals(SNR, x, h, noise)

        h_hat = pra.experimental.deconvolve(y, x, length=h_len)
        rmse = np.sqrt(np.linalg.norm(h_hat - h)**2 / h_len)

        print('rmse=', rmse, '(tol=', tol, ')')

        self.assertTrue(rmse < tol)

    def test_wiener_deconvolve_hann_noiseless(self):

        h = h_hann
        h_len = h_hann.shape[0]
        SNR = 1000.
        tol = 1e-7

        y, sigma_noise = generate_signals(SNR, x, h, noise)

        h_hat = pra.experimental.wiener_deconvolve(y, x, length=h_len, noise_variance=sigma_noise**2)
        rmse = np.sqrt(np.linalg.norm(h_hat - h)**2 / h_len)

        print('rmse=', rmse, '(tol=', tol, ')')

        self.assertTrue(rmse < tol)

if __name__ == '__main__':

    import matplotlib.pyplot as plt

    h = h_hann
    y, sigma_noise = generate_signals(SNR, x, h, noise)

    h_hat1 = pra.experimental.deconvolve(y, x, length=h_len)
    res1 = np.linalg.norm(y - fftconvolve(x, h_hat1))**2 / y.shape[0]
    mse1 = np.linalg.norm(h_hat1 - h)**2 / h_len

    h_hat2 = pra.experimental.wiener_deconvolve(y, x, length=h_len, noise_variance=sigma_noise**2, let_n_points=15)
    res2 = np.linalg.norm(y - fftconvolve(x, h_hat2))**2 / y.shape[0]
    mse2 = np.linalg.norm(h_hat2 - h)**2 / h_len

    print('MSE naive: rmse=', np.sqrt(mse1), ' res=', pra.dB(res1, power=True))
    print('MSE Wiener: rmse=', np.sqrt(mse2), ' res=', pra.dB(res1, power=True))

    plt.plot(h)
    plt.plot(h_hat1)
    plt.plot(h_hat2)
    plt.legend(['Original', 'Naive', 'Wiener'])
    plt.show()



from __future__ import division, print_function
import scipy.linalg as la
import numpy as np


class AdaptiveFilter:
    '''
    The dummy base class of an adaptive filter. This class doesn't compute
    anything. It merely stores values in a buffer. It is used as a template
    for all other algorithms.
    '''
    
    def __init__(self, length):
        
        # filter length
        self.length = length
        
        self.reset()
    
    def reset(self):
        '''
        Reset the state of the adaptive filter
        '''
        
        # index
        self.n = 0
        
        # filter
        self.w = np.zeros((self.length))
        
        # system input signal
        self.x = np.zeros((self.length))
        
        # reference signal
        self.d = np.zeros((1))
        
    def update(self, x_n, d_n):
        '''
        Updates the adaptive filter with a new sample

        Parameters
        ----------
        x_n: float
            the new input sample
        d_n: float
            the new noisy reference signal
        '''
        
        self.n += 1
        
        # update buffers
        self.x[1:] = self.x[0:-1]
        self.x[0] = x_n
        self.d = d_n
        
    def name(self):
        
        return self.__class__.__name__




# Copyright (c) 2019 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Test code for RT60 measurement routine
"""
import numpy as np
import pyroomacoustics as pra

eps = 1e-15

room = pra.ShoeBox([10, 7, 6], fs=16000, absorption=0.35, max_order=17)
room.add_source([3, 2.5, 1.7])
room.add_microphone_array(pra.MicrophoneArray(np.array([[7, 3.7, 1.1]]).T, room.fs))
room.compute_rir()

ir = room.rir[0][0]


def test_rt60():
    """
    Very basic test that runs the function and checks that the value
    returned with different sampling frequencies are correct.
    """

    t60_samples = pra.experimental.measure_rt60(ir)
    t60_s = pra.experimental.measure_rt60(ir, fs=room.fs)

    assert abs(t60_s - t60_samples / room.fs) < eps

    t30_samples = pra.experimental.measure_rt60(ir, decay_db=30)
    t30_s = pra.experimental.measure_rt60(ir, decay_db=30, fs=room.fs)

    assert abs(t30_s - t30_samples / room.fs) < eps


def test_rt60_plot():
    """
    Simple run of the plot without any output.

    Check for runtime errors only.
    """

    import matplotlib
    matplotlib.use('Agg')

    pra.experimental.measure_rt60(ir, plot=True)
    pra.experimental.measure_rt60(ir, plot=True, rt60_tgt=0.3)


if __name__ == "__main__":

    test_rt60()
    test_rt60_plot()



"""
Block based processing
======================

This subpackage contains routines for continuous *realtime-like* block
processing of signals with the STFT. The routines are written to take
advantage of fast FFT libraries like `pyfftw` or `mkl` when available.

DFT
    | A class for performing DFT of real signals
    | :py:obj:`pyroomacoustics.transform.dft`
STFT
    | A class for continuous STFT processing and frequency domain filtering
    | :py:obj:`pyroomacoustics.transform.stft`

"""


from .dft import DFT
from .stft import STFT
from . import stft


def analysis(*args, **kwargs):
    import warnings
    warnings.warn(
        "The `pyroomacoustics.transform.analysis` function is deprecated in favor of "
        "`pyroomacoustics.transform.stft.analysis`.",
        DeprecationWarning,
    )
    return stft.analysis(*args, **kwargs)


def synthesis(*args, **kwargs):
    import warnings
    warnings.warn(
        "The `pyroomacoustics.transform.synthesis` function is deprecated in favor of "
        "'pyroomacoustics.transform.stft.synthesis`.",
        DeprecationWarning,
    )
    return stft.synthesis(*args, **kwargs)


def compute_synthesis_window(*args, **kwargs):
    import warnings
    warnings.warn(
        "The `pyroomacoustics.transform.compute_synthesis_window` function is "
        "deprecated in favor of "
        "`pyroomacoustics.transform.stft.compute_synthesis_window`.",
        DeprecationWarning,
    )
    return stft.compute_synthesis_window(*args, **kwargs)



# Subband Least Mean Squares
# Copyright (C) 2019  Eric Bezzam
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

import numpy as np


class SubbandLMS:

    '''
    Frequency domain implementation of LMS. Adaptive filter for each
    subband.

    Parameters
    ----------
    num_taps: int 
        length of the filter
    num_bands: int 
        number of frequency bands, i.e. number of filters
    mu: float, optional
        step size for each subband (default 0.5)
    nlms: bool, optional
        whether or not to normalize as in NLMS (default is True)
    '''

    def __init__(self, num_taps, num_bands, mu=0.5, nlms=True):

        self.num_taps = num_taps
        self.num_bands = num_bands
        self.mu = mu
        self.nlms = nlms

        self.reset()

    def reset(self):

        # filter bank
        self.W = np.zeros((self.num_taps, self.num_bands), dtype=np.complex64)

        # input signal
        self.X = np.zeros((self.num_taps, self.num_bands), dtype=np.complex64)

        # reference signal
        self.D = np.zeros(self.num_bands, dtype=np.complex64)

        # error signal
        self.E = np.zeros(self.num_bands, dtype=np.complex64)

    def update(self, X_n, D_n):
        '''
        Updates the adaptive filters for each subband with the new
        block of input data.

        Parameters
        ----------
        X_n: numpy array, float
            new input signal (to unknown system) in frequency domain
        D_n: numpy array, float
            new noisy reference signal in frequency domain
        '''

        # update buffers
        self.X[1:, :] = self.X[0:-1, :]
        self.X[0, :] = X_n
        self.D = D_n

        # a priori error
        self.E = self.D - np.diag(np.dot(hermitian(self.W), self.X))

        # compute update
        update = self.mu * np.tile(self.E.conj(), (self.num_taps, 1)) * self.X
        if self.nlms:
            update /= np.tile(np.diag(np.dot(hermitian(self.X), self.X)),
                              (self.num_taps, 1)) + 1e-6

        # update filter coefficients
        self.W += update


def hermitian(X):
        '''
        Compute and return Hermitian transpose
        '''
        return X.conj().T





# Test of the geometry routines
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux, Sidney Barthe
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import unittest
import numpy as np

import pyroomacoustics as pra

class TestGeometryRoutines(unittest.TestCase):

    def test_area_square(self):
        self.assertEqual(pra.libroom.area_2d_polygon([[0, 4, 4, 0], [0, 0, 4, 4]]), 4*4)

    def test_area_triangle(self):
        self.assertEqual(pra.libroom.area_2d_polygon([[0, 4, 2], [0, 0, 2]]), (4*2)/2)

    def test_ccw3p_counterclockwise(self):
        self.assertEqual(pra.libroom.ccw3p([0, 0], [2, 0], [1, 1]), 1)

    def test_ccw3p_clockwise(self):
        self.assertEqual(pra.libroom.ccw3p([1, 1], [2, 0], [0, 0]), -1)

    def test_ccw3p_collinear(self):
        self.assertEqual(pra.libroom.ccw3p([0, 0], [1, 0], [2, 0]), 0)

    def test_intersection2DSegments_cross(self):
        loc = np.zeros(2, dtype=np.float32)
        ret = pra.libroom.intersection_2d_segments([-2, 0], [2, 0], [0, -2], [0, 2], loc)
        i = np.allclose(loc, np.zeros(2))
        self.assertTrue(all([i, ret == 0]))  # intersection is valid

    def test_intersection2DSegments_T(self):
        loc = np.zeros(2, dtype=np.float32)
        ret = pra.libroom.intersection_2d_segments([0, 0], [2, 0], [2, -2], [2, 2], loc)
        i = np.allclose(loc, np.r_[2,0])
        self.assertTrue(all([i, ret == 1]))  # intersection on endpoint of 1st seg

    def test_intersection2DSegments_T2(self):
        loc = np.zeros(2, dtype=np.float32)
        ret = pra.libroom.intersection_2d_segments([-2, 0], [2, 0], [0, -2], [0, 0], loc)
        i = np.allclose(loc, np.r_[0,0])
        self.assertTrue(all([i, ret == 2]))

    def test_intersection2DSegments_L(self):
        loc = np.zeros(2, dtype=np.float32)
        ret = pra.libroom.intersection_2d_segments([0, 0], [2, 0], [2, 2], [2, 0], loc)
        i = np.allclose(loc, np.r_[2,0])
        self.assertTrue(all([i, ret == 3]))

    def test_intersection2DSegments_parallel(self):
        loc = np.zeros(2, dtype=np.float32)
        ret = pra.libroom.intersection_2d_segments([0, 0], [2, 0], [0, 1], [2, 1], loc)
        self.assertTrue(ret == -1)

    def test_intersection2DSegments_notTouching(self):
        p = np.zeros(2, dtype=np.float32)
        ret = pra.libroom.intersection_2d_segments([0, 0], [2, 0], [1, 4], [1, 1], p)
        self.assertTrue(ret == -1)

    def test_intersectionSegmentPlane_through(self):
        p = np.zeros(3, dtype=np.float32)
        ret = pra.libroom.intersection_3d_segment_plane([2, 2, 2], [2, 2, -2], [2, 2, 0], [0, 0, -1], p)
        i = np.allclose(p, np.r_[2,2,0])
        self.assertTrue(all([i, ret == 0]))

    def test_intersectionSegmentPlane_touching(self):
        p = np.zeros(3, dtype=np.float32)
        ret = pra.libroom.intersection_3d_segment_plane([2, 2, 2], [2, 2, 0], [2, 2, 0], [0, 0, -1], p)
        i = np.allclose(p, np.r_[2,2,0])
        self.assertTrue(all([i, ret == 1]))

    def test_intersectionSegmentPlane_notTouching(self):
        p = np.zeros(3, dtype=np.float32)
        ret = pra.libroom.intersection_3d_segment_plane([2, 2, 2], [2, 2, 1], [2, 2, 0], [0, 0, -1], p)
        self.assertTrue(all([ret == -1]))

    def test_intersectionSegmentPlane_inside(self):
        p = np.zeros(3, dtype=np.float32)
        ret = pra.libroom.intersection_3d_segment_plane([0, 2, 0], [2, 2, 0], [2, 2, 0], [0, 0, -1], p)
        self.assertTrue(all([ret == -1]))

    def test_intersectionSegmentPolygonSurface_through(self):
        wall = pra.libroom.Wall([[0, 4, 4, 0], [0, 0, 4, 4], [0, 0, 0, 0]])
        p = np.zeros(3, dtype=np.float32)
        ret = wall.intersection([2,2,2], [2,2,-2], p)
        i = np.allclose(p, np.r_[2,2,0])
        self.assertTrue(all([i, ret == pra.libroom.Wall.Isect.VALID]))

    def test_intersectionSegmentPolygonSurface_touching(self):
        wall = pra.libroom.Wall([[0, 4, 4, 0], [0, 0, 4, 4], [0, 0, 0, 0]])
        p = np.zeros(3, dtype=np.float32)
        ret = wall.intersection([2,2,2], [2,2,0], p)
        i = np.allclose(p, np.r_[2,2,0])
        self.assertTrue(all([i, ret == pra.libroom.Wall.Isect.ENDPT]))

    def test_intersectionSegmentPolygonSurface_border(self):
        wall = pra.libroom.Wall([[0, 4, 4, 0], [0, 0, 4, 4], [0, 0, 0, 0]])
        p = np.zeros(3, dtype=np.float32)
        ret = wall.intersection([0,0,2], [0,0,-2], p)
        i = np.allclose(p, np.r_[0,0,0])
        self.assertTrue(all([i, ret == pra.libroom.Wall.Isect.BNDRY]))

    def test_intersectionSegmentPolygonSurface_miss(self):
        wall = pra.libroom.Wall([[0, 4, 4, 0], [0, 0, 4, 4], [0, 0, 0, 0]])
        p = np.zeros(3, dtype=np.float32)
        ret = wall.intersection([-1,-1,2], [-1,-1,-2], p)
        self.assertTrue(all([ret == pra.libroom.Wall.Isect.NONE]))

    def test_intersectionSegmentPolygonSurface_miss_2(self):
        corners = np.array([
            [0, 0, 0, 0],
            [3, 0, 0, 3],
            [0, 0, 4, 4]
            ])
        wall = pra.libroom.Wall(corners)
        p = np.array([ 2. ,  3. ,  1.7])
        p0 = np.array([-2.348,  3.27 ,  8.91 ])
        loc = np.zeros(p.shape[0], dtype=np.float32)
        ret = wall.intersection(p, p0, loc)
        self.assertTrue(ret == pra.libroom.Wall.Isect.NONE)

    def test_isInside2DPolygon_inside1(self):
        p = np.r_[0, -2]
        corners = np.array([
            [ 0,     -2,           0,  2,],
            [-0,     -2, -np.sqrt(8), -2,],
            ])
        ret = pra.libroom.is_inside_2d_polygon(p, corners)
        self.assertTrue(ret == 0)
        
    def test_isInside2DPolygon_inside(self):
        ret = pra.libroom.is_inside_2d_polygon([2, 2], [[0, 4, 4, 0], [0, 0, 4, 4]])
        self.assertTrue(all([ret == 0]))
        
    def test_isInside2DPolygon_onBorder(self):
        ret = pra.libroom.is_inside_2d_polygon([0, 2], [[0, 4, 4, 0], [0, 0, 4, 4]])
        self.assertTrue(all([ret == 1]))
        
    def test_isInside2DPolygon_onCorner(self):
        ret = pra.libroom.is_inside_2d_polygon([4, 4], [[0, 4, 4, 0], [0, 0, 4, 4]])
        self.assertTrue(all([ret == 1]))
        
    def test_isInside2DPolygon_outside(self):
        ret = pra.libroom.is_inside_2d_polygon([5, 5], [[0, 4, 4, 0], [0, 0, 4, 4]])
        self.assertTrue(all([ret == -1]))

    def test_isInside2DPolygon_parallel_wall(self):
        corners = [[0, 0, 1, 1, 3, 3], [0, 1, 1, 2, 2, 0]]
        ret = pra.libroom.is_inside_2d_polygon([2, 1], corners)
        self.assertTrue(all([ret == 0]))

    def test_isInside2DPolygon_at_corner(self):
        corners = [[0, 1, 1, 3, 3], [0, 1, 2, 2, 0]]
        ret = pra.libroom.is_inside_2d_polygon([2, 1], corners)
        self.assertTrue(all([ret == 0]))

    def test_isInside2DPolygon_just_out(self):
        corners = [[0, 1, 1, 3, 3], [0, 1, 2, 2, 0]]
        ret = pra.libroom.is_inside_2d_polygon([4, 2], corners)
        self.assertTrue(all([ret == -1]))

    def test_cosangleBetween2D(self):
        eps = 0.001
        result = pra.libroom.cos_angle_between([1.,0.],[-1.,-1.])
        self.assertTrue(np.allclose(result, -np.sqrt(2)/2., atol=eps))

    def test_cosangleBetween3D_orthog(self):
        eps = 0.001
        result = pra.libroom.cos_angle_between([1.,0.,0.],[0.,0.,1.])
        self.assertTrue(np.allclose(result, 0, atol=eps))

    def test_cosangleBetween3D_colinear(self):
        eps = 0.001
        result = pra.libroom.cos_angle_between([1.,1.,0.],[-1.,-1.,0.])
        self.assertTrue(np.allclose(result, -1, atol=eps))

    def test_cosangleBetween3D_special(self):
        eps = 0.001
        result = pra.libroom.cos_angle_between([36.,1.,-4.],[12.,-1.,2.])
        self.assertTrue(np.allclose(result,0.956345613,  atol=eps))


    def test_dist_line_point2D(self):

        start = [3,3]
        end = [6,9]
        point = [11,4]

        eps = 0.001
        res = pra.libroom.dist_line_point(start, end, point)
        self.assertTrue(abs(res - np.sqrt(6*6+3*3)) < eps)

    def test_dist_line_point2D_vert(self):

        start = [-4,12]
        end = [-4,27]
        point = [7,10]

        eps = 0.001
        res = pra.libroom.dist_line_point(start, end, point)
        self.assertTrue(abs(res - 11) < eps)

    def test_dist_line_point3D(self):

        start = [0,0,0]
        end = [1,2,3]
        point = [4,5,6]

        eps = 0.001
        res = pra.libroom.dist_line_point(start, end, point)
        self.assertTrue(abs(res - 1.963961012) < eps)

    def test_dist_line_point3D_online(self):

        start = [0,0,0]
        end = [0,0,3]
        point = [0,0,6]

        eps = 0.001
        res = pra.libroom.dist_line_point(start, end, point)
        self.assertTrue(abs(res) < eps)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pyroomacoustics as pra

sig = np.arange(10)
room_size = [10, 9, 8]
source_loc0 = [1.5, 1.7, 2.1]
source_loc1 = [3.5, 7.7, 2.1]
mic0 = [7, 8, 3.9]
mic1 = [7.87, 3.6, 6.1]


def test_add_source_mic():

    room = pra.ShoeBox(room_size).add_source(source_loc0).add_microphone(mic0)

    assert len(room.sources) == 1
    assert np.allclose(room.sources[0].position, source_loc0)
    assert len(room.mic_array) == 1
    assert room.mic_array.R.shape == (3, 1)
    assert np.allclose(room.mic_array.R[:, 0], mic0)

    room.add_microphone(mic1).add_source(source_loc1)

    assert len(room.sources) == 2
    assert np.allclose(room.sources[1].position, source_loc1)
    assert len(room.mic_array) == 2
    assert np.allclose(room.mic_array.R[:, 0], mic0)
    assert np.allclose(room.mic_array.R[:, 1], mic1)
    assert room.mic_array.R.shape == (3, 2)


def test_add_source_mic_obj():

    room = pra.ShoeBox(room_size)

    source0 = pra.SoundSource(source_loc0, signal=sig)
    source1 = pra.SoundSource(source_loc1, signal=sig)

    mic_array0 = pra.MicrophoneArray(np.c_[mic0], fs=room.fs)
    mic_array1 = pra.MicrophoneArray(np.c_[mic1], fs=room.fs)

    room.add(source0).add(mic_array0)

    assert len(room.sources) == 1
    assert np.allclose(room.sources[0].position, source_loc0)
    assert len(room.mic_array) == 1
    assert room.mic_array.R.shape == (3, 1)
    assert np.allclose(room.mic_array.R[:, 0], mic0)

    room.add(mic_array1).add(source1)

    assert len(room.sources) == 2
    assert np.allclose(room.sources[1].position, source_loc1)
    assert len(room.mic_array) == 2
    assert np.allclose(room.mic_array.R[:, 0], mic0)
    assert np.allclose(room.mic_array.R[:, 1], mic1)
    assert room.mic_array.R.shape == (3, 2)


def test_add_source_mic_obj_2():

    room = pra.ShoeBox(room_size)

    source0 = pra.SoundSource(source_loc0, signal=sig)
    source1 = pra.SoundSource(source_loc1, signal=sig)
    mic_array = pra.MicrophoneArray(np.c_[mic0, mic1], fs=room.fs)

    room.add(source0).add(source1).add(mic_array)

    assert len(room.sources) == 2
    assert np.allclose(room.sources[0].position, source_loc0)
    assert np.allclose(room.sources[1].position, source_loc1)
    assert len(room.mic_array) == 2
    assert np.allclose(room.mic_array.R[:, 0], mic0)
    assert np.allclose(room.mic_array.R[:, 1], mic1)
    assert room.mic_array.R.shape == (3, 2)


def test_add_source_mic_ndarray():

    source0 = pra.SoundSource(source_loc0, signal=sig)
    source1 = pra.SoundSource(source_loc1, signal=sig)
    mic_array = np.c_[mic0, mic1]

    room = (
        pra.ShoeBox(room_size).add(source0).add(source1).add_microphone_array(mic_array)
    )

    assert len(room.sources) == 2
    assert np.allclose(room.sources[0].position, source_loc0)
    assert np.allclose(room.sources[1].position, source_loc1)
    assert len(room.mic_array) == 2
    assert np.allclose(room.mic_array.R[:, 0], mic0)
    assert np.allclose(room.mic_array.R[:, 1], mic1)
    assert room.mic_array.R.shape == (3, 2)


def test_add_source_mic_ndarray_2():

    source0 = pra.SoundSource(source_loc0, signal=sig)
    source1 = pra.SoundSource(source_loc1, signal=sig)
    mic_array = np.c_[mic0, mic1]

    room = pra.ShoeBox(room_size).add(source0).add(source1).add_microphone(mic_array)

    assert len(room.sources) == 2
    assert np.allclose(room.sources[0].position, source_loc0)
    assert np.allclose(room.sources[1].position, source_loc1)
    assert len(room.mic_array) == 2
    assert np.allclose(room.mic_array.R[:, 0], mic0)
    assert np.allclose(room.mic_array.R[:, 1], mic1)
    assert room.mic_array.R.shape == (3, 2)


if __name__ == "__main__":
    test_add_source_mic()
    test_add_source_mic_obj()
    test_add_source_mic_obj_2()
    test_add_source_mic_ndarray()
    test_add_source_mic_ndarray_2()




from __future__ import division, print_function
from unittest import TestCase
import numpy as np
from scipy.signal import fftconvolve
import pyroomacoustics as pra
from pyroomacoustics.transform import STFT

'''
We create a signal, a simple filter and compute their convolution.

Then we test STFT block procesing with and without overlap,
and with and without filtering. Simulating a case of real-time
block processing.
'''

# test parameters
tol = 5e-6
np.random.seed(0)
D = 4
transform = 'numpy'   # 'numpy', 'pyfftw', or 'mkl'

# filter to apply
h_len = 99
h = np.ones((h_len, D))
h /= np.linalg.norm(h, axis=0)

# test signal (noise)
x = np.random.randn(100000, D).astype(np.float32)

# convolved signal
y = np.zeros((x.shape[0] + h_len - 1, x.shape[1]))
for i in range(x.shape[1]):
    y[:,i] = fftconvolve(x[:,i], h[:,i])


def incorrect_input_size(D, num_frames):

    if D == 1:
        x_local = x[:,0]
    else:
        x_local = x[:,:D]

    # parameters
    block_size = 512
    hop = block_size

    # create STFT object
    stft = STFT(block_size, hop=hop,
        channels=D, 
        transform=transform,
        num_frames=num_frames)

    try:  # passing more frames than 'hop'
        stft.analysis(x_local)
        computed = False
    except:
        computed = True

    return computed



def no_overlap_no_filter(D, num_frames=1, fixed_memory=False,
                        streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at
                    a time
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:,0]
    else:
        x_local = x[:,:D]

    # parameters
    block_size = 512  # make sure the FFT size is a power of 2
    hop = block_size  # no overlap
    if not streaming:
        num_samples = (num_frames-1)*hop+block_size
        x_local = x_local[:num_samples,]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(block_size, hop=hop, channels=D, 
                transform=transform, num_frames=num_frames, streaming=streaming)
    else:
        stft = STFT(block_size, hop=hop, channels=D, 
            transform=transform, streaming=streaming)

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if streaming:

        n = 0
        hop_frames = hop*num_frames
        # process the signals while full blocks are available
        while  x_local.shape[0] - n > hop_frames:
            stft.analysis(x_local[n:n+hop_frames,])
            processed_x[n:n+hop_frames,] = stft.synthesis()
            n += hop_frames

    else:

        stft.analysis(x_local)
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

    error = np.max(np.abs(x_local[:n,] - processed_x[:n,]))

    return error

def with_arbitrary_overlap_synthesis_window(D, num_frames=1, fixed_memory=False,
                        streaming=True, overlap=0.5):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at
                    a time
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:,0]
    else:
        x_local = x[:,:D]

    # parameters
    block_size = 512  # make sure the FFT size is a power of 2
    hop = int((1 - overlap) * block_size)  # quarter overlap
    if not streaming:
        num_samples = (num_frames-1)*hop+block_size
        x_local = x_local[:num_samples,]

    analysis_window = pra.hann(block_size)
    synthesis_window = pra.transform.compute_synthesis_window(analysis_window, hop)

    # Create the STFT object
    if fixed_memory:
        stft = STFT(block_size, hop=hop, channels=D,
                transform=transform, num_frames=num_frames,
                analysis_window=analysis_window, synthesis_window=synthesis_window,
                streaming=streaming)
    else:
        stft = STFT(block_size, hop=hop, channels=D,
                analysis_window=analysis_window, synthesis_window=synthesis_window,
                transform=transform, streaming=streaming)

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if streaming:

        n = 0
        hop_frames = hop*num_frames
        # process the signals while full blocks are available
        while  x_local.shape[0] - n > hop_frames:
            stft.analysis(x_local[n:n+hop_frames,])
            processed_x[n:n+hop_frames,] = stft.synthesis()
            n += hop_frames

        error = np.max(np.abs(x_local[:n-block_size+hop,] - processed_x[block_size-hop:n,]))

        if 20 * np.log10(error) > -10:
            import matplotlib.pyplot as plt
            if x_local.ndim == 1:
                plt.plot(x_local[:n-block_size+hop])
                plt.plot(processed_x[block_size-hop:n])
            else:
                plt.plot(x_local[:n-block_size+hop,0])
                plt.plot(processed_x[block_size-hop:n,0])
            plt.show()

    else:

        stft.analysis(x_local)
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

        L = block_size - hop
        error = np.max(np.abs(x_local[L:-L,] - processed_x[L:,]))

        if 20 * np.log10(error) > -10:
            import matplotlib.pyplot as plt
            if x_local.ndim == 1:
                plt.plot(x_local[L:-L])
                plt.plot(processed_x[L:])
            else:
                plt.plot(x_local[L:-L,0])
                plt.plot(processed_x[L:,0])
            plt.show()


    return error


def no_overlap_with_filter(D, num_frames=1, fixed_memory=False,
                        streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at 
                    a time 
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:,0]
        y_local = y[:,0]
        h_local = h[:,0]
    else:
        x_local = x[:,:D]
        y_local = y[:,:D]
        h_local = h[:,:D]

    # parameters
    block_size = 512 - h_len + 1  # make sure the FFT size is a power of 2
    hop = block_size  # no overlap
    if not streaming:
        num_samples = (num_frames-1)*hop+block_size
        x_local = x_local[:num_samples,]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(block_size, hop=hop, channels=D, 
                transform=transform, num_frames=num_frames, streaming=streaming)
    else:
        stft = STFT(block_size, hop=hop, channels=D, 
            transform=transform, streaming=streaming)
    
    # setup the filter
    stft.set_filter(h_local, zb=h_len - 1)

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if not streaming:

        stft.analysis(x_local)
        stft.process()
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

    else:

        n = 0
        hop_frames = hop*num_frames
        # process the signals while full blocks are available
        while  x_local.shape[0] - n > hop_frames:
            stft.analysis(x_local[n:n+hop_frames,])
            stft.process()  # apply the filter
            processed_x[n:n+hop_frames,] = stft.synthesis()
            n += hop_frames

    error = np.max(np.abs(y_local[:n,] - processed_x[:n,]))

    return error


def with_half_overlap_no_filter(D, num_frames=1, fixed_memory=False,
                        streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at 
                    a time 
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:,0]
    else:
        x_local = x[:,:D]

    # parameters
    block_size = 512  # make sure the FFT size is a power of 2
    hop = block_size // 2  # half overlap
    window = pra.hann(block_size)  # the analysis window
    if not streaming:
        num_samples = (num_frames-1)*hop+block_size
        x_local = x_local[:num_samples,]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(block_size, hop=hop, channels=D, 
                transform=transform, num_frames=num_frames, 
                analysis_window=window, streaming=streaming)
    else:
        stft = STFT(block_size, hop=hop, channels=D, 
            transform=transform, analysis_window=window, streaming=streaming)

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if not streaming:

        stft.analysis(x_local)
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

        error = np.max(np.abs(x_local[block_size-hop:n-hop,] 
            - processed_x[block_size-hop:n-hop,]))

    else:

        n = 0
        hop_frames = hop*num_frames
        # process the signals while full blocks are available
        while  x_local.shape[0] - n > hop_frames:
            stft.analysis(x_local[n:n+hop_frames,])
            processed_x[n:n+hop_frames,] = stft.synthesis()
            n += hop_frames

        error = np.max(np.abs(x_local[:n-hop,] - processed_x[hop:n,]))

    return error

def with_half_overlap_with_filter(D, num_frames=1, fixed_memory=False,
                        streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at 
                    a time 
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:,0]
        y_local = y[:,0]
        h_local = h[:,0]
    else:
        x_local = x[:,:D]
        y_local = y[:,:D]
        h_local = h[:,:D]

    # parameters
    block_size = 512 - h_len + 1  # make sure the FFT size is a power of 2
    hop = block_size // 2  # half overlap
    window = pra.hann(block_size)  # the analysis window
    if not streaming:
        num_samples = (num_frames-1)*hop+block_size
        x_local = x_local[:num_samples,]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(block_size, hop=hop, channels=D, 
                transform=transform, num_frames=num_frames, 
                analysis_window=window, streaming=streaming)
    else:
        stft = STFT(block_size, hop=hop, channels=D, 
            transform=transform, analysis_window=window, streaming=streaming)

    # setup the filter
    stft.set_filter(h_local, zb=h_len - 1)

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if not streaming:

        stft.analysis(x_local)
        stft.process()
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

        error = np.max(np.abs(y_local[block_size:n-block_size,] 
            - processed_x[block_size:n-block_size,]))

    else:

        n = 0
        hop_frames = hop*num_frames
        # process the signals while full blocks are available
        while  x_local.shape[0] - n > hop_frames:
            stft.analysis(x_local[n:n+hop_frames,])
            stft.process()  # apply the filter
            processed_x[n:n+hop_frames,] = stft.synthesis()
            n += hop_frames

        error = np.max(np.abs(y_local[:n-hop,] - processed_x[hop:n,]))

        # if D==1:
        #     import matplotlib.pyplot as plt
        #     plt.figure()
        #     plt.plot(y_local)
        #     plt.plot(processed_x)
        #     plt.show()

    return error


def call_all_stft_tests(num_frames=1, fixed_memory=False, streaming=True, 
    overlap=True):

    error = no_overlap_no_filter(1, num_frames, fixed_memory,
        streaming)
    print('no overlap, no filter, mono             : %0.0f dB' 
        % (20*np.log10(error)))

    error = no_overlap_no_filter(D, num_frames, fixed_memory, streaming)
    print('no overlap, no filter, multichannel     : %0.0f dB' 
        % (20*np.log10(error)))

    error = no_overlap_with_filter(1, num_frames, fixed_memory,
        streaming)
    print('no overlap, with filter, mono           : %0.0f dB' 
        % (20*np.log10(error)))

    error = no_overlap_with_filter(D, num_frames, fixed_memory,
        streaming)
    print('no overlap, with filter, multichannel   : %0.0f dB' 
        % (20*np.log10(error)))

    if overlap:
        error = with_half_overlap_no_filter(1, num_frames, fixed_memory,
            streaming)
        print('half overlap, no filter, mono           : %0.0f dB' 
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(1, num_frames,
                fixed_memory, streaming, overlap=0.5)
        print('half overlap, no filter, with synthesis windows, mono : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(D, num_frames,
                fixed_memory, streaming, overlap=0.5)
        print('half overlap, no filter, with synthesis windows, multichannel : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(1, num_frames,
                fixed_memory, streaming, overlap=0.75)
        print('3/4 overlap, no filter, with synthesis windows, mono : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(D, num_frames,
                fixed_memory, streaming, overlap=0.75)
        print('3/4 overlap, no filter, with synthesis windows, multichannel : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(1, num_frames,
                fixed_memory, streaming, overlap=0.84)
        print('84/100 overlap, no filter, with synthesis windows, mono : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(D, num_frames,
                fixed_memory, streaming, overlap=0.84)
        print('84/100 overlap, no filter, with synthesis windows, multichannel : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(1, num_frames,
                fixed_memory, streaming, overlap=0.26)
        print('26/100 overlap, no filter, with synthesis windows, mono : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(D, num_frames,
                fixed_memory, streaming, overlap=0.26)
        print('26/100 overlap, no filter, with synthesis windows, multichannel : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(1, num_frames,
                fixed_memory, streaming, overlap=7/8)
        print('7/8 overlap, no filter, with synthesis windows, mono : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(D, num_frames,
                fixed_memory, streaming, overlap=7/8)
        print('7/8 overlap, no filter, with synthesis windows, multichannel : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(1, num_frames,
                fixed_memory, streaming, overlap=0.25)
        print('1/4 overlap, no filter, with synthesis windows, mono : %0.0f dB'
            % (20*np.log10(error)))

        error = with_arbitrary_overlap_synthesis_window(D, num_frames,
                fixed_memory, streaming, overlap=0.25)
        print('1/4 overlap, no filter, with synthesis windows, multichannel : %0.0f dB'
            % (20*np.log10(error)))

        error = with_half_overlap_no_filter(D, num_frames, fixed_memory,
            streaming)
        print('half overlap, no filter, multichannel   : %0.0f dB'
            % (20*np.log10(error)))

        error = with_half_overlap_with_filter(1, num_frames,
            fixed_memory, streaming)
        print('half overlap, with filter, mono         : %0.0f dB'
            % (20*np.log10(error)))

        error = with_half_overlap_with_filter(D, num_frames,
            fixed_memory, streaming)
        print('half overlap, with filter, multichannel : %0.0f dB'
            % (20*np.log10(error)))

        error = with_half_overlap_with_filter(D, num_frames,
            fixed_memory, streaming)
        print('half overlap, with filter, multichannel : %0.0f dB'
            % (20*np.log10(error)))
    print()



class TestSTFT(TestCase):

    def test_incorrect_input_check(self):
        result = incorrect_input_size(1, 100)
        self.assertTrue(result)
        result = incorrect_input_size(D, 100)
        self.assertTrue(result)

    def test_no_overlap_no_filter_mono(self):
        error = no_overlap_no_filter(D=1, num_frames=1, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=1, num_frames=50, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=1, num_frames=1, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=1, num_frames=50, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=1, num_frames=1, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=1, num_frames=50, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=1, num_frames=1, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=1, num_frames=50, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)


    def test_no_overlap_no_filter_multichannel(self):
        error = no_overlap_no_filter(D=D, num_frames=1, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=D, num_frames=50, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=D, num_frames=1, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=D, num_frames=50, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=D, num_frames=1, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=D, num_frames=50, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=D, num_frames=1, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(D=D, num_frames=50, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)

    def test_no_overlap_with_filter_mono(self):
        error = no_overlap_with_filter(D=1, num_frames=1, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=1, num_frames=50, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=1, num_frames=1, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=1, num_frames=50, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=1, num_frames=1, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=1, num_frames=50, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=1, num_frames=1, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=1, num_frames=50, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)

    def test_no_overlap_with_filter_multichannel(self):
        error = no_overlap_with_filter(D=D, num_frames=1, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=D, num_frames=50, fixed_memory=False, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=D, num_frames=1, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=D, num_frames=50, fixed_memory=True, 
            streaming=True)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=D, num_frames=1, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=D, num_frames=50, fixed_memory=False, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=D, num_frames=1, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(D=D, num_frames=50, fixed_memory=True, 
            streaming=False)
        self.assertTrue(error < tol)


    def test_with_half_overlap_no_filter_mono(self):
        error = with_half_overlap_no_filter(D=1, num_frames=1, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=1, num_frames=50, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=1, num_frames=1, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=1, num_frames=50, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=1, num_frames=50, 
            fixed_memory=False, streaming=False)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=1, num_frames=50, 
            fixed_memory=True, streaming=False)
        self.assertTrue(error < tol)


    def test_with_half_overlap_no_filter_multichannel(self):
        error = with_half_overlap_no_filter(D=D, num_frames=1, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=D, num_frames=50, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=D, num_frames=1, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=D, num_frames=50, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=D, num_frames=50, 
            fixed_memory=False, streaming=False)
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(D=D, num_frames=50, 
            fixed_memory=True, streaming=False)
        self.assertTrue(error < tol)

    def test_with_half_overlap_with_filter_mono(self):
        error = with_half_overlap_with_filter(D=1, num_frames=1, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=1, num_frames=50, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=1, num_frames=1, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=1, num_frames=50, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=1, num_frames=50, 
            fixed_memory=False, streaming=False)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=1, num_frames=50, 
            fixed_memory=True, streaming=False)
        self.assertTrue(error < tol)

    def test_with_half_overlap_with_filter_multichannel(self):
        error = with_half_overlap_with_filter(D=D, num_frames=1, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=D, num_frames=50, 
            fixed_memory=False, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=D, num_frames=1, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=D, num_frames=50, 
            fixed_memory=True, streaming=True)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=D, num_frames=50, 
            fixed_memory=False, streaming=False)
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(D=D, num_frames=50, 
            fixed_memory=True, streaming=False)
        self.assertTrue(error < tol)

    def test_with_arbitrary_overlap_synthesis_window_multichannel(self):
        overlaps = [0.5, 0.75, 0.84, 0.26, 7/8, 0.25]
        for overlap in overlaps:
            error = with_arbitrary_overlap_synthesis_window(D, num_frames=1,
                    fixed_memory=False, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(D, num_frames=50,
                    fixed_memory=False, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(D, num_frames=1,
                    fixed_memory=True, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(D, num_frames=50,
                    fixed_memory=True, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(D, num_frames=50,
                    fixed_memory=False, streaming=False, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(D, num_frames=50,
                    fixed_memory=True, streaming=False, overlap=overlap)
            self.assertTrue(error < tol)

    def test_with_arbitrary_overlap_synthesis_window_mono(self):
        overlaps = [0.5, 0.75, 0.84, 0.26, 7/8, 0.25]
        for overlap in overlaps:
            error = with_arbitrary_overlap_synthesis_window(1, num_frames=1,
                    fixed_memory=False, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(1, num_frames=50,
                    fixed_memory=False, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(1, num_frames=1,
                    fixed_memory=True, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(1, num_frames=50,
                    fixed_memory=True, streaming=True, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(1, num_frames=50,
                    fixed_memory=False, streaming=False, overlap=overlap)
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(1, num_frames=50,
                    fixed_memory=True, streaming=False, overlap=overlap)
            self.assertTrue(error < tol)


if __name__ == "__main__":

    print()
    print("TEST INFO")
    print("-------------------------------------------------------------")
    print("Max error in dB for randomly generated signal of %d samples."
        % len(x))
    print("Multichannel corresponds to %d channels." % D)
    print("-------------------------------------------------------------")
    print()

    print("---ONE FRAME, STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(num_frames=1, fixed_memory=False, streaming=True)

    print("---MULTIPLE FRAMES, STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(num_frames=50, fixed_memory=False, streaming=True)

    print("---ONE FRAME, STREAMING, FIXED MEMORY")
    num_frames = 1
    result = incorrect_input_size(1, num_frames)
    print('incorrect input size, mono              :', result)
    result = incorrect_input_size(D, num_frames)
    print('incorrect input size, multichannel      :', result)
    call_all_stft_tests(num_frames=num_frames, fixed_memory=True, 
        streaming=True)

    print("---MULTIPLE FRAME, STREAMING, FIXED MEMORY")
    num_frames=50
    result = incorrect_input_size(1, num_frames)
    print('incorrect input size, mono              :', result)
    result = incorrect_input_size(D, num_frames)
    print('incorrect input size, multichannel      :', result)
    call_all_stft_tests(num_frames=num_frames, fixed_memory=True, 
        streaming=True)

    print("---ONE FRAME, NON-STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(num_frames=1, fixed_memory=False, streaming=False, 
        overlap=False)

    print("---MULTIPLE FRAMES, NON-STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(num_frames=50, fixed_memory=False, streaming=False)

    print("---ONE FRAME, NON-STREAMING, FIXED MEMORY")
    call_all_stft_tests(num_frames=1, fixed_memory=True, streaming=False, 
        overlap=False)

    print("---MULTIPLE FRAMES, NON-STREAMING, FIXED MEMORY")
    call_all_stft_tests(num_frames=50, fixed_memory=True, streaming=False)







import numpy as np
import pyroomacoustics as pra

def test_soundsource_basic():

    the_signal = np.ones(10)
    source = pra.SoundSource([1., 1.], signal=the_signal)

    the_signal_2 = np.zeros(10)
    source.add_signal(the_signal_2)

if __name__ == '__main__':
    
    test_soundsource_basic()




# Various Beamforming Methods
# Copyright (C) 2019  Robin Scheibler, Sidney Barthe, Ivan Dokmanic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

from __future__ import division

import numpy as np
import scipy.linalg as la

from .parameters import constants
from . import utilities as u
from .soundsource import build_rir_matrix
from . import windows
from . import transform


# =========================================================================
# Free (non-class-member) functions related to beamformer design
# =========================================================================


def H(A, **kwargs):
    """Returns the conjugate (Hermitian) transpose of a matrix."""

    return np.transpose(A, **kwargs).conj()


def sumcols(A):
    """
    Sums the columns of a matrix (np.array).

    The output is a 2D np.array
    of dimensions M x 1.
    """

    return np.sum(A, axis=1, keepdims=1)


def mdot(*args):
    """
    Left-to-right associative matrix multiplication of multiple 2D ndarrays.
    """

    ret = args[0]
    for a in args[1:]:
        ret = np.dot(ret, a)

    return ret


def distance(x, y):
    """
    Computes the distance matrix E.

    E[i,j] = sqrt(sum((x[:,i]-y[:,j])**2)).
    x and y are DxN ndarray containing N D-dimensional vectors.
    """

    # Assume x, y are arrays, *not* matrices
    x = np.array(x)
    y = np.array(y)

    # return np.sqrt((x[0,:,np.newaxis]-y[0,:])**2 +
    # (x[1,:,np.newaxis]-y[1,:])**2)

    return np.sqrt(np.sum((x[:, :, np.newaxis] - y[:, np.newaxis, :]) ** 2, axis=0))


def unit_vec2D(phi):
    return np.array([[np.cos(phi), np.sin(phi)]]).T


def linear_2D_array(center, M, phi, d):
    """
    Creates an array of uniformly spaced linear points in 2D

    Parameters
    ----------
    center: array_like
        The center of the array
    M: int
        The number of points
    phi: float
        The counterclockwise rotation of the array (from the x-axis)
    d: float
        The distance between neighboring points

    Returns
    -------
    ndarray (2, M)
        The array of points
    """
    u = unit_vec2D(phi)
    return (
        np.array(center)[:, np.newaxis]
        + d * (np.arange(M)[np.newaxis, :] - (M - 1.0) / 2.0) * u
    )


def circular_2D_array(center, M, phi0, radius):
    """
    Creates an array of uniformly spaced circular points in 2D

    Parameters
    ----------
    center: array_like
        The center of the array
    M: int
        The number of points
    phi0: float
        The counterclockwise rotation of the first element in the array (from
        the x-axis)
    radius: float
        The radius of the array

    Returns
    -------
    ndarray (2, M)
        The array of points
    """
    phi = np.arange(M) * 2.0 * np.pi / M
    return np.array(center)[:, np.newaxis] + radius * np.vstack(
        (np.cos(phi + phi0), np.sin(phi + phi0))
    )


def poisson_2D_array(center, M, d):
    """
    Create array of 2D positions drawn from Poisson process.

    Parameters
    ----------
    center: array_like
        The center of the array
    M: int
        The number of points in the first dimension
    M: int
        The number of points in the second dimension
    phi: float
        The counterclockwise rotation of the array (from the x-axis)
    d: float
        The distance between neighboring points

    Returns
    -------
    ndarray (2, M * N)
        The array of points
    """

    from numpy.random import standard_exponential, randint

    R = d * standard_exponential((2, M)) * (2 * randint(0, 2, (2, M)) - 1)
    R = R.cumsum(axis=1)
    R -= R.mean(axis=1)[:, np.newaxis]
    R += np.array([center]).T

    return R


def square_2D_array(center, M, N, phi, d):
    """
    Creates an array of uniformly spaced grid points in 2D

    Parameters
    ----------
    center: array_like
        The center of the array
    M: int
        The number of points in the first dimension
    M: int
        The number of points in the second dimension
    phi: float
        The counterclockwise rotation of the array (from the x-axis)
    d: float
        The distance between neighboring points

    Returns
    -------
    ndarray (2, M * N)
        The array of points
    """

    c = linear_2D_array(center, M, phi + np.pi / 2.0, d)
    R = np.zeros((2, M * N))
    for i in np.arange(M):
        R[:, i * N : (i + 1) * N] = linear_2D_array(c[:, i], N, phi, d)

    return R


def spiral_2D_array(center, M, radius=1.0, divi=3, angle=None):
    """
    Generate an array of points placed on a spiral

    Parameters
    ----------

    center: array_like
        location of the center of the array
    M: int
        number of microphones
    radius: float
        microphones are contained within a cirle of this radius (default 1)
    divi: int
        number of rotations of the spiral (default 3)
    angle: float
        the angle offset of the spiral (default random)

    Returns
    -------
    ndarray (2, M * N)
        The array of points
    """
    num_seg = int(np.ceil(M / divi))

    pos_array_norm = np.linspace(0, radius, num=M, endpoint=False)

    pos_array_angle = (
        np.reshape(
            np.tile(np.pi * 2 * np.arange(divi) / divi, num_seg), (divi, -1), order="F"
        )
        + np.linspace(0, 2 * np.pi / divi, num=num_seg, endpoint=False)[np.newaxis, :]
    )
    pos_array_angle = np.insert(pos_array_angle.flatten("F")[: M - 1], 0, 0)

    if angle is None:
        pos_array_angle += np.random.rand() * np.pi / divi
    else:
        pos_array_angle += angle

    pos_mic_x = pos_array_norm * np.cos(pos_array_angle)
    pos_mic_y = pos_array_norm * np.sin(pos_array_angle)

    return np.array([pos_mic_x, pos_mic_y])


def fir_approximation_ls(weights, T, n1, n2):

    freqs_plus = np.array(weights.keys())[:, np.newaxis]
    freqs = np.vstack([freqs_plus, -freqs_plus])
    omega = 2 * np.pi * freqs
    omega_discrete = omega * T

    n = np.arange(n1, n2)

    # Create the DTFT transform matrix corresponding to a discrete set of
    # frequencies and the FIR filter indices
    F = np.exp(-1j * omega_discrete * n)

    w_plus = np.array(weights.values())[:, :, 0]
    w = np.vstack([w_plus, w_plus.conj()])

    return np.linalg.pinv(F).dot(w)


# =========================================================================
# Classes (microphone array and beamformer related)
# =========================================================================


class MicrophoneArray(object):

    """Microphone array class."""

    def __init__(self, R, fs):

        R = np.array(R)
        self.dim = R.shape[0]  # are we in 2D or in 3D

        # Check the shape of the passed array
        if self.dim != 2 and self.dim != 3:
            dim_mismatch = True
        else:
            dim_mismatch = False

        if R.ndim != 2 or dim_mismatch:
            raise ValueError(
                "The location of microphones should be described by an array_like "
                "object with 2 dimensions of shape `(2 or 3, n_mics)` "
                "where `n_mics` is the number of microphones. Each column contains "
                "the location of a microphone."
            )

        self.R = R  # array geometry

        self.fs = fs  # sampling frequency of microphones

        self.signals = None

        self.center = np.mean(R, axis=1, keepdims=True)

    def record(self, signals, fs):
        """
        This simulates the recording of the signals by the microphones.
        In particular, if the microphones and the room simulation
        do not use the same sampling frequency, down/up-sampling
        is done here.

        Parameters
        ----------

        signals:
            An ndarray with as many lines as there are microphones.
        fs:
            the sampling frequency of the signals.
        """

        if signals.shape[0] != self.M:
            raise NameError(
                "The signals array should have as many lines as "
                "there are microphones."
            )

        if signals.ndim != 2:
            raise NameError("The signals should be a 2D array.")

        if fs != self.fs:
            try:
                import samplerate

                fs_ratio = self.fs / float(fs)
                newL = int(fs_ratio * signals.shape[1]) - 1
                self.signals = np.zeros((self.M, newL))
                # samplerate resample function considers columns as channels
                # (hence the transpose)
                for m in range(self.M):
                    self.signals[m] = samplerate.resample(
                        signals[m], fs_ratio, "sinc_best"
                    )
            except ImportError:
                raise ImportError(
                    "The samplerate package must be installed for"
                    " resampling of the signals."
                )

        else:
            self.signals = signals

    def to_wav(self, filename, mono=False, norm=False, bitdepth=np.float):
        """
        Save all the signals to wav files.

        Parameters
        ----------
        filename: str
            the name of the file
        mono: bool, optional
            if true, records only the center channel floor(M / 2) (default
            `False`)
        norm: bool, optional
            if true, normalize the signal to fit in the dynamic range (default
            `False`)
        bitdepth: int, optional
            the format of output samples [np.int8/16/32/64 or np.float
            (default)]
        """
        from scipy.io import wavfile

        if mono is True:
            signal = self.signals[self.M // 2]
        else:
            signal = self.signals.T  # each column is a channel

        float_types = [float, np.float, np.float32, np.float64]

        if bitdepth in float_types:
            bits = None
        elif bitdepth is np.int8:
            bits = 8
        elif bitdepth is np.int16:
            bits = 16
        elif bitdepth is np.int32:
            bits = 32
        elif bitdepth is np.int64:
            bits = 64
        else:
            raise NameError("No such type.")

        if norm:
            from .utilities import normalize

            signal = normalize(signal, bits=bits)

        signal = np.array(signal, dtype=bitdepth)

        wavfile.write(filename, self.fs, signal)

    def append(self, locs):
        """
        Add some microphones to the array

        Parameters
        ----------
        locs: numpy.ndarray (2 or 3, n_mics)
            Adds `n_mics` microphones to the array. The coordinates are passed as
            a `numpy.ndarray` with each column containing the coordinates of a
            microphone.
        """
        if isinstance(locs, MicrophoneArray):
            self.R = np.concatenate((self.R, locs.R), axis=1)
        else:
            self.R = np.concatenate((self.R, locs), axis=1)

        # in case there was already some signal recorded, just pad with zeros
        if self.signals is not None:
            self.signals = np.concatenate(
                (
                    self.signals,
                    np.zeros(
                        (locs.shape[1], self.signals.shape[1]), dtype=self.signals.dtype
                    ),
                ),
                axis=0,
            )

    def __len__(self):
        return self.R.shape[1]

    @property
    def M(self):
        return self.__len__()


class Beamformer(MicrophoneArray):

    """
    At some point, in some nice way, the design methods
    should also go here. Probably with generic arguments.

    Parameters
    ----------
    R: numpy.ndarray
        Mics positions
    fs: int
        Sampling frequency
    N: int, optional
        Length of FFT, i.e. number of FD beamforming weights, equally spaced.
        Defaults to 1024.
    Lg: int, optional
        Length of time-domain filters. Default to N.
    hop: int, optional
        Hop length for frequency domain processing. Default to N/2.
    zpf: int, optional
        Front zero padding length for frequency domain processing. Default is 0.
    zpb: int, optional
        Zero padding length for frequency domain processing. Default is 0.
    """

    def __init__(self, R, fs, N=1024, Lg=None, hop=None, zpf=0, zpb=0):
        MicrophoneArray.__init__(self, R, fs)

        # only support even length (in freq)
        if N % 2 == 1:
            N += 1

        self.N = int(N)  # FFT length

        if Lg is None:
            self.Lg = N  # TD filters length
        else:
            self.Lg = int(Lg)

        # setup lengths for FD processing
        self.zpf = int(zpf)
        self.zpb = int(zpb)
        self.L = self.N - self.zpf - self.zpb
        if hop is None:
            self.hop = self.L // 2
        else:
            self.hop = hop

        # for now only support equally spaced frequencies
        self.frequencies = np.arange(0, self.N // 2 + 1) / self.N * float(self.fs)

        # weights will be computed later, the array is of shape (M, N/2+1)
        self.weights = None

        # the TD beamforming filters (M, Lg)
        self.filters = None

    def __add__(self, y):
        """ Concatenates two beamformers together."""

        newR = np.concatenate((self.R, y.R), axis=1)
        return Beamformer(
            newR, self.fs, self.Lg, self.N, hop=self.hop, zpf=self.zpf, zpb=self.zpb
        )

    def filters_from_weights(self, non_causal=0.0):
        """
        Compute time-domain filters from frequency domain weights.

        Parameters
        ----------
        non_causal: float, optional
            ratio of filter coefficients used for non-causal part
        """

        if self.weights is None:
            raise NameError("Weights must be defined.")

        self.filters = np.zeros((self.M, self.Lg))

        if self.N <= self.Lg:

            # go back to time domain and shift DC to center
            tw = np.fft.irfft(np.conj(self.weights), axis=1, n=self.N)
            self.filters[:, : self.N] = np.concatenate(
                (tw[:, -self.N // 2 :], tw[:, : self.N // 2]), axis=1
            )

        elif self.N > self.Lg:

            # Least-square projection
            for i in np.arange(self.M):
                Lgp = np.floor((1 - non_causal) * self.Lg)
                Lgm = self.Lg - Lgp
                # the beamforming weights in frequency are the complex
                # conjugates of the FT of the filter
                w = np.concatenate((np.conj(self.weights[i]), self.weights[i, -2:0:-1]))

                # create partial Fourier matrix
                k = np.arange(self.N)[:, np.newaxis]
                l = np.concatenate((np.arange(self.N - Lgm, self.N), np.arange(Lgp)))
                F = np.exp(-2j * np.pi * k * l / self.N)

                self.filters[i] = np.real(np.linalg.lstsq(F, w, rcond=None)[0])

    def weights_from_filters(self):

        if self.filters is None:
            raise NameError("Filters must be defined.")

        # this is what we want to use, really.
        # self.weights = np.conj(np.fft.rfft(self.filters, n=self.N, axis=1))

        # quick hack to be able to use MKL acceleration package from anaconda
        self.weights = np.zeros((self.M, self.N // 2 + 1), dtype=np.complex128)
        for m in range(self.M):
            self.weights[m] = np.conj(np.fft.rfft(self.filters[m], n=self.N))

    def steering_vector_2D(self, frequency, phi, dist, attn=False):

        phi = np.array([phi]).reshape(phi.size)

        # Assume phi and dist are measured from the array's center
        X = dist * np.array([np.cos(phi), np.sin(phi)]) + self.center

        D = distance(self.R, X)
        omega = 2 * np.pi * frequency

        if attn:
            # TO DO 1: This will mean slightly different absolute value for
            # every entry, even within the same steering vector. Perhaps a
            # better paradigm is far-field with phase carrier.
            return 1.0 / (4 * np.pi) / D * np.exp(-1j * omega * D / constants.get("c"))
        else:
            return np.exp(-1j * omega * D / constants.get("c"))

    def steering_vector_2D_from_point(self, frequency, source, attn=True, ff=False):
        """ Creates a steering vector for a particular frequency and source

        Args:
            frequency
            source: location in cartesian coordinates
            attn: include attenuation factor if True
            ff:   uses far-field distance if true

        Return:
            A 2x1 ndarray containing the steering vector.
        """

        X = np.array(source)
        if X.ndim == 1:
            X = source[:, np.newaxis]

        omega = 2 * np.pi * frequency

        # normalize for far-field if requested
        if ff:
            # unit vectors pointing towards sources
            p = X - self.center
            p /= np.linalg.norm(p)

            # The projected microphone distances on the unit vectors
            D = -1 * np.dot(self.R.T, p)

            # subtract minimum in each column
            D -= np.min(D)

        else:

            D = distance(self.R, X)

        phase = np.exp(-1j * omega * D / constants.get("c"))

        if attn:
            # TO DO 1: This will mean slightly different absolute value for
            # every entry, even within the same steering vector. Perhaps a
            # better paradigm is far-field with phase carrier.
            return 1.0 / (4 * np.pi) / D * phase
        else:
            return phase

    def response(self, phi_list, frequency):

        i_freq = np.argmin(np.abs(self.frequencies - frequency))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError(
                "Beamforming weights or filters need to be computed" " first."
            )

        # For the moment assume that we are in 2D
        bfresp = np.dot(
            H(self.weights[:, i_freq]),
            self.steering_vector_2D(
                self.frequencies[i_freq], phi_list, constants.get("ffdist")
            ),
        )

        return self.frequencies[i_freq], bfresp

    def response_from_point(self, x, frequency):

        i_freq = np.argmin(np.abs(self.frequencies - frequency))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError(
                "Beamforming weights or filters need to be computed" " first."
            )

        # For the moment assume that we are in 2D
        bfresp = np.dot(
            H(self.weights[:, i_freq]),
            self.steering_vector_2D_from_point(
                self.frequencies[i_freq], x, attn=True, ff=False
            ),
        )

        return self.frequencies[i_freq], bfresp

    def plot_response_from_point(self, x, legend=None):

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError(
                "Beamforming weights or filters need to be computed" " first."
            )

        if x.ndim == 0:
            x = np.array([x])

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        HF = np.zeros((x.shape[1], self.frequencies.shape[0]), dtype=complex)
        for k, p in enumerate(x.T):
            for i, f in enumerate(self.frequencies):
                r = np.dot(
                    H(self.weights[:, i]),
                    self.steering_vector_2D_from_point(f, p, attn=True, ff=False),
                )
                HF[k, i] = r[0]

        plt.subplot(2, 1, 1)
        plt.title("Beamformer response")
        for hf in HF:
            plt.plot(self.frequencies, np.abs(hf))
        plt.ylabel("Modulus")
        plt.axis("tight")
        plt.legend(legend)

        plt.subplot(2, 1, 2)
        for hf in HF:
            plt.plot(self.frequencies, np.unwrap(np.angle(hf)))
        plt.ylabel("Phase")
        plt.xlabel("Frequency [Hz]")
        plt.axis("tight")
        plt.legend(legend)

    def plot_beam_response(self):

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError(
                "Beamforming weights or filters need to be computed" " first."
            )

        phi = np.linspace(-np.pi, np.pi - np.pi / 180, 360)
        freq = self.frequencies

        resp = np.zeros((freq.shape[0], phi.shape[0]), dtype=complex)

        for i, f in enumerate(freq):
            # For the moment assume that we are in 2D
            resp[i, :] = np.dot(
                H(self.weights[:, i]),
                self.steering_vector_2D(f, phi, constants.get("ffdist")),
            )

        H_abs = np.abs(resp) ** 2
        H_abs /= H_abs.max()
        H_abs = 10 * np.log10(H_abs + 1e-10)

        p_min = 0
        p_max = 100
        vmin, vmax = np.percentile(H_abs.flatten(), [p_min, p_max])

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        plt.imshow(
            H_abs,
            aspect="auto",
            origin="lower",
            interpolation="sinc",
            vmax=vmax,
            vmin=vmin,
        )

        plt.xlabel("Angle [rad]")
        xticks = [-np.pi, -np.pi / 2, 0, np.pi / 2, np.pi]
        for i, p in enumerate(xticks):
            xticks[i] = np.argmin(np.abs(p - phi))
        xticklabels = ["$-\pi$", "$-\pi/2$", "0", "$\pi/2$", "$\pi$"]
        plt.setp(plt.gca(), "xticks", xticks)
        plt.setp(plt.gca(), "xticklabels", xticklabels)

        plt.ylabel("Freq [kHz]")
        yticks = np.zeros(4)
        f_0 = np.floor(self.fs / 8000.0)
        for i in np.arange(1, 5):
            yticks[i - 1] = np.argmin(np.abs(freq - 1000.0 * i * f_0))
        # yticks = np.array(plt.getp(plt.gca(), 'yticks'), dtype=np.int)
        plt.setp(plt.gca(), "yticks", yticks)
        plt.setp(plt.gca(), "yticklabels", np.arange(1, 5) * f_0)

    def snr(self, source, interferer, f, R_n=None, dB=False):

        i_f = np.argmin(np.abs(self.frequencies - f))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError(
                "Beamforming weights or filters need to be computed" " first."
            )

        # This works at a single frequency because otherwise we need to pass
        # many many covariance matrices. Easy to change though (you can also
        # have frequency independent R_n).

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        # To compute the SNR, we /must/ use the real steering vectors, so no
        # far field, and attn=True
        A_good = self.steering_vector_2D_from_point(
            self.frequencies[i_f], source.images, attn=True, ff=False
        )

        if interferer is not None:
            A_bad = self.steering_vector_2D_from_point(
                self.frequencies[i_f], interferer.images, attn=True, ff=False
            )
            R_nq = R_n + sumcols(A_bad) * H(sumcols(A_bad))
        else:
            R_nq = R_n

        w = self.weights[:, i_f]
        a_1 = sumcols(A_good)

        SNR = np.real(mdot(H(w), a_1, H(a_1), w) / mdot(H(w), R_nq, w))

        if dB is True:
            SNR = 10 * np.log10(SNR)

        return SNR

    def udr(self, source, interferer, f, R_n=None, dB=False):

        i_f = np.argmin(np.abs(self.frequencies - f))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError(
                "Beamforming weights or filters need to be computed" " first."
            )

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        A_good = self.steering_vector_2D_from_point(
            self.frequencies[i_f], source.images, attn=True, ff=False
        )

        if interferer is not None:
            A_bad = self.steering_vector_2D_from_point(
                self.frequencies[i_f], interferer.images, attn=True, ff=False
            )
            R_nq = R_n + sumcols(A_bad).dot(H(sumcols(A_bad)))
        else:
            R_nq = R_n

        w = self.weights[:, i_f]

        UDR = np.real(mdot(H(w), A_good, H(A_good), w) / mdot(H(w), R_nq, w))
        if dB is True:
            UDR = 10 * np.log10(UDR)

        return UDR

    def process(self, FD=False):

        if self.signals is None or len(self.signals) == 0:
            raise NameError("No signal to beamform.")

        if FD is True:

            # STFT processing
            if self.weights is None and self.filters is not None:
                self.weights_from_filters()
            elif self.weights is None and self.filters is None:
                raise NameError(
                    "Beamforming weights or filters need to be " "computed first."
                )

            # create window functions
            analysis_win = windows.hann(self.L)

            # perform STFT
            sig_stft = transform.analysis(
                self.signals.T,
                L=self.L,
                hop=self.hop,
                win=analysis_win,
                zp_back=self.zpb,
                zp_front=self.zpf,
            )

            # beamform
            sig_stft_bf = np.sum(sig_stft * self.weights.conj().T, axis=2)

            # back to time domain
            output = transform.synthesis(
                sig_stft_bf, L=self.L, hop=self.hop, zp_back=self.zpb, zp_front=self.zpf
            )

            # remove the zero padding from output signal
            if self.zpb == 0:
                output = output[self.zpf :]
            else:
                output = output[self.zpf : -self.zpb]

        else:

            # TD processing

            if self.weights is not None and self.filters is None:
                self.filters_from_weights()
            elif self.weights is None and self.filters is None:
                raise NameError(
                    "Beamforming weights or filters need to be " "computed first."
                )

            from scipy.signal import fftconvolve

            # do real STFT of first signal
            output = fftconvolve(self.filters[0], self.signals[0])
            for i in range(1, len(self.signals)):
                output += fftconvolve(self.filters[i], self.signals[i])

        return output

    def plot(self, sum_ir=False, FD=True):

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is not None and self.filters is None:
            self.filters_from_weights()
        elif self.weights is None and self.filters is None:
            raise NameError(
                "Beamforming weights or filters need to be " "computed first."
            )

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        if FD is True:
            plt.subplot(2, 2, 1)
            plt.plot(self.frequencies, np.abs(self.weights.T))
            plt.title("Beamforming weights [modulus]")
            plt.xlabel("Frequency [Hz]")
            plt.ylabel("Weight modulus")

            plt.subplot(2, 2, 2)
            plt.plot(self.frequencies, np.unwrap(np.angle(self.weights.T), axis=0))
            plt.title("Beamforming weights [phase]")
            plt.xlabel("Frequency [Hz]")
            plt.ylabel("Unwrapped phase")

            plt.subplot(2, 1, 2)

        plt.plot(np.arange(self.Lg) / float(self.fs), self.filters.T)

        plt.title("Beamforming filters")
        plt.xlabel("Time [s]")
        plt.ylabel("Filter amplitude")
        plt.axis("tight")
        plt.tight_layout(pad=0.1)

    def far_field_weights(self, phi):
        """
        This method computes weight for a far field at infinity

        phi: direction of beam
        """

        u = unit_vec2D(phi)
        proj = np.dot(u.T, self.R - self.center)[0]

        # normalize the first arriving signal to ensure a causal filter
        proj -= proj.max()

        self.weights = np.exp(
            2j * np.pi * self.frequencies[:, np.newaxis] * proj / constants.get("c")
        ).T

    def rake_delay_and_sum_weights(
        self, source, interferer=None, R_n=None, attn=True, ff=False
    ):

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        K = source.images.shape[1] - 1

        for i, f in enumerate(self.frequencies):
            W = self.steering_vector_2D_from_point(f, source.images, attn=attn, ff=ff)
            self.weights[:, i] = 1.0 / self.M / (K + 1) * np.sum(W, axis=1)

    def rake_one_forcing_weights(
        self, source, interferer=None, R_n=None, ff=False, attn=True
    ):

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        for i, f in enumerate(self.frequencies):
            if interferer is None:
                A_bad = np.array([[]])
            else:
                A_bad = self.steering_vector_2D_from_point(
                    f, interferer.images, attn=attn, ff=ff
                )

            R_nq = R_n + sumcols(A_bad).dot(H(sumcols(A_bad)))

            A_s = self.steering_vector_2D_from_point(f, source.images, attn=attn, ff=ff)
            R_nq_inv = np.linalg.pinv(R_nq)
            D = np.linalg.pinv(mdot(H(A_s), R_nq_inv, A_s))

            self.weights[:, i] = sumcols(mdot(R_nq_inv, A_s, D))[:, 0]

    def rake_max_sinr_weights(
        self, source, interferer=None, R_n=None, rcond=0.0, ff=False, attn=True
    ):
        """
        This method computes a beamformer focusing on a number of specific
        sources and ignoring a number of interferers.

        INPUTS
          * source     : source locations
          * interferer : interferer locations
        """

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        for i, f in enumerate(self.frequencies):

            A_good = self.steering_vector_2D_from_point(
                f, source.images, attn=attn, ff=ff
            )

            if interferer is None:
                A_bad = np.array([[]])
            else:
                A_bad = self.steering_vector_2D_from_point(
                    f, interferer.images, attn=attn, ff=ff
                )

            a_good = sumcols(A_good)
            a_bad = sumcols(A_bad)

            # TO DO: Fix this (check for numerical rank, use the low rank
            # approximation)
            K_inv = np.linalg.pinv(
                a_bad.dot(H(a_bad)) + R_n + rcond * np.eye(A_bad.shape[0])
            )
            self.weights[:, i] = (K_inv.dot(a_good) / mdot(H(a_good), K_inv, a_good))[
                :, 0
            ]

    def rake_max_udr_weights(
        self, source, interferer=None, R_n=None, ff=False, attn=True
    ):

        if source.images.shape[1] == 1:
            self.rake_max_sinr_weights(
                source.images, interferer.images, R_n=R_n, ff=ff, attn=attn
            )
            return

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        for i, f in enumerate(self.frequencies):
            A_good = self.steering_vector_2D_from_point(
                f, source.images, attn=attn, ff=ff
            )

            if interferer is None:
                A_bad = np.array([[]])
            else:
                A_bad = self.steering_vector_2D_from_point(
                    f, interferer.images, attn=attn, ff=ff
                )

            R_nq = R_n + sumcols(A_bad).dot(H(sumcols(A_bad)))

            C = np.linalg.cholesky(R_nq)
            l, v = np.linalg.eig(
                mdot(np.linalg.inv(C), A_good, H(A_good), H(np.linalg.inv(C)))
            )

            self.weights[:, i] = np.linalg.inv(H(C)).dot(v[:, 0])

    def rake_max_udr_filters(
        self, source, interferer=None, R_n=None, delay=0.03, epsilon=5e-3
    ):
        """
        Compute directly the time-domain filters maximizing the
        Useful-to-Detrimental Ratio (UDR).

        This beamformer is not practical. It maximizes the UDR ratio in the time
        domain directly without imposing flat response towards the source of
        interest. This results in severe distortion of the desired signal.

        Parameters
        ----------
        source: pyroomacoustics.SoundSource
            the desired source
        interferer: pyroomacoustics.SoundSource, optional
            the interfering source
        R_n: ndarray, optional
            the noise covariance matrix, it should be (M * Lg)x(M * Lg) where M
            is the number of sensors and Lg the filter length
        delay: float, optional
            the signal delay introduced by the beamformer (default 0.03 s)
        epsilon: float
        """
        if delay > self.Lg / self.fs:
            print("Warning: filter length shorter than beamformer delay")

        if R_n is None:
            R_n = np.zeros((self.M * self.Lg, self.M * self.Lg))

        if interferer is not None:
            H = build_rir_matrix(
                self.R,
                (source, interferer),
                self.Lg,
                self.fs,
                epsilon=epsilon,
                unit_damping=True,
            )
            L = H.shape[1] // 2
        else:
            H = build_rir_matrix(
                self.R, (source,), self.Lg, self.fs, epsilon=epsilon, unit_damping=True
            )
            L = H.shape[1]

        # Delay of the system in samples
        kappa = int(delay * self.fs)
        precedence = int(0.030 * self.fs)

        # the constraint
        n = int(np.minimum(L, kappa + precedence))
        Hnc = H[:, :kappa]
        Hpr = H[:, kappa:n]
        A = np.dot(Hpr, Hpr.T)
        B = np.dot(Hnc, Hnc.T) + np.dot(H[:, L:], H[:, L:].T) + R_n

        if interferer is not None:
            Hc = H[:, n:L]
            B += np.dot(Hc, Hc.T)

        # solve the problem
        SINR, v = la.eigh(
            A,
            b=B,
            eigvals=(self.M * self.Lg - 1, self.M * self.Lg - 1),
            overwrite_a=True,
            overwrite_b=True,
            check_finite=False,
        )
        g_val = np.real(v[:, 0])

        # reshape and store
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        return SINR[0]

    def rake_perceptual_filters(
        self, source, interferer=None, R_n=None, delay=0.03, d_relax=0.035, epsilon=5e-3
    ):
        """
        Compute directly the time-domain filters for a perceptually motivated
        beamformer. The beamformer minimizes noise and interference, but relaxes
        the response of the filter within the 30 ms following the delay.
        """

        if delay > self.Lg / self.fs:
            print("Warning: filter length shorter than beamformer delay")

        if R_n is None:
            R_n = np.zeros((self.M * self.Lg, self.M * self.Lg))

        # build the channel matrix
        if interferer is not None:
            H = build_rir_matrix(
                self.R,
                (source, interferer),
                self.Lg,
                self.fs,
                epsilon=epsilon,
                unit_damping=True,
            )
            L = H.shape[1] // 2
        else:
            H = build_rir_matrix(
                self.R, (source,), self.Lg, self.fs, epsilon=epsilon, unit_damping=True
            )
            L = H.shape[1]

        # Delay of the system in samples
        tau = int(delay * self.fs)
        kappa = int(d_relax * self.fs)

        # the constraint
        A = np.concatenate((H[:, : tau + 1], H[:, tau + kappa :]), axis=1)
        b = np.zeros((A.shape[1], 1))
        b[tau, 0] = 1

        # We first assume the sample are uncorrelated
        K_nq = R_n
        if interferer is not None:
            K_nq += np.dot(H[:, L:], H[:, L:].T)

        # causal response construction
        C = la.cho_factor(K_nq, overwrite_a=True, check_finite=False)
        B = la.cho_solve(C, A)
        D = np.dot(A.T, B)
        C = la.cho_factor(D, overwrite_a=True, check_finite=False)
        x = la.cho_solve(C, b)
        g_val = np.dot(B, x)

        # reshape and store
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        A = np.dot(g_val.T, H[:, :L])
        num = np.dot(A, A.T)
        denom = np.dot(np.dot(g_val.T, K_nq), g_val)

        return num / denom

    def rake_max_sinr_filters(self, source, interferer, R_n, epsilon=5e-3, delay=0.0):
        """
        Compute the time-domain filters of SINR maximizing beamformer.
        """

        H = build_rir_matrix(
            self.R,
            (source, interferer),
            self.Lg,
            self.fs,
            epsilon=epsilon,
            unit_damping=True,
        )
        L = H.shape[1] / 2

        # We first assume the sample are uncorrelated
        K_s = np.dot(H[:, :L], H[:, :L].T)
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # Compute TD filters using generalized Rayleigh coefficient maximization
        SINR, v = la.eigh(
            K_s,
            b=K_nq,
            eigvals=(self.M * self.Lg - 1, self.M * self.Lg - 1),
            overwrite_a=True,
            overwrite_b=True,
            check_finite=False,
        )
        g_val = np.real(v[:, 0])

        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        return SINR[0]

    def rake_distortionless_filters(
        self, source, interferer, R_n, delay=0.03, epsilon=5e-3
    ):
        """
        Compute time-domain filters of a beamformer minimizing noise and
        interference while forcing a distortionless response towards the source.
        """

        H = build_rir_matrix(
            self.R,
            (source, interferer),
            self.Lg,
            self.fs,
            epsilon=epsilon,
            unit_damping=True,
        )
        L = H.shape[1] / 2

        # We first assume the sample are uncorrelated
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # constraint
        kappa = int(delay * self.fs)
        A = H[:, :L]
        b = np.zeros((L, 1))
        b[kappa, 0] = 1

        # filter computation
        C = la.cho_factor(K_nq, overwrite_a=True, check_finite=False)
        B = la.cho_solve(C, A)
        D = np.dot(A.T, B)
        C = la.cho_factor(D, overwrite_a=True, check_finite=False)
        x = la.cho_solve(C, b)
        g_val = np.dot(B, x)

        # reshape and store
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        A = np.dot(g_val.T, H[:, :L])
        num = np.dot(A, A.T)
        denom = np.dot(np.dot(g_val.T, K_nq), g_val)

        return num / denom

    def rake_mvdr_filters(self, source, interferer, R_n, delay=0.03, epsilon=5e-3):
        """
        Compute the time-domain filters of the minimum variance distortionless
        response beamformer.
        """

        H = build_rir_matrix(
            self.R,
            (source, interferer),
            self.Lg,
            self.fs,
            epsilon=epsilon,
            unit_damping=True,
        )
        L = H.shape[1] // 2

        # the constraint vector
        kappa = int(delay * self.fs)
        h = H[:, kappa]

        # We first assume the sample are uncorrelated
        R_xx = np.dot(H[:, :L], H[:, :L].T)
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # Compute the TD filters
        C = la.cho_factor(R_xx + K_nq, check_finite=False)
        g_val = la.cho_solve(C, h)

        g_val /= np.inner(h, g_val)
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        num = np.inner(g_val.T, np.dot(R_xx, g_val))
        denom = np.inner(np.dot(g_val.T, K_nq), g_val)

        return num / denom

    def rake_one_forcing_filters(self, sources, interferers, R_n, epsilon=5e-3):
        """
        Compute the time-domain filters of a beamformer with unit response
        towards multiple sources.
        """

        dist_mat = distance(self.R, sources.images)
        s_time = dist_mat / constants.get("c")
        s_dmp = 1.0 / (4 * np.pi * dist_mat)

        dist_mat = distance(self.R, interferers.images)
        i_time = dist_mat / constants.get("c")
        i_dmp = 1.0 / (4 * np.pi * dist_mat)

        # compute offset needed for decay of sinc by epsilon
        offset = np.maximum(s_dmp.max(), i_dmp.max()) / (np.pi * self.fs * epsilon)
        t_min = np.minimum(s_time.min(), i_time.min())
        t_max = np.maximum(s_time.max(), i_time.max())

        # adjust timing
        s_time -= t_min - offset
        i_time -= t_min - offset
        Lh = np.ceil((t_max - t_min + 2 * offset) * float(self.fs))

        # the channel matrix
        K = sources.images.shape[1]
        Lg = self.Lg
        off = (Lg - Lh) / 2
        L = self.Lg + Lh - 1

        H = np.zeros((Lg * self.M, 2 * L))
        As = np.zeros((Lg * self.M, K))

        for r in np.arange(self.M):

            # build constraint matrix
            hs = u.low_pass_dirac(
                s_time[r, :, np.newaxis], s_dmp[r, :, np.newaxis], self.fs, Lh
            )[:, ::-1]
            As[r * Lg + off : r * Lg + Lh + off, :] = hs.T

            # build interferer RIR matrix
            hx = u.low_pass_dirac(
                s_time[r, :, np.newaxis], s_dmp[r, :, np.newaxis], self.fs, Lh
            ).sum(axis=0)
            H[r * Lg : (r + 1) * Lg, :L] = u.convmtx(hx, Lg).T

            # build interferer RIR matrix
            hq = u.low_pass_dirac(
                i_time[r, :, np.newaxis], i_dmp[r, :, np.newaxis], self.fs, Lh
            ).sum(axis=0)
            H[r * Lg : (r + 1) * Lg, L:] = u.convmtx(hq, Lg).T

        ones = np.ones((K, 1))

        # We first assume the sample are uncorrelated
        K_x = np.dot(H[:, :L], H[:, :L].T)
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # Compute the TD filters
        K_nq_inv = np.linalg.inv(K_x + K_nq)
        C = np.dot(K_nq_inv, As)
        B = np.linalg.inv(np.dot(As.T, C))
        g_val = np.dot(C, np.dot(B, ones))
        self.filters = g_val.reshape((self.M, Lg))

        # compute and return SNR
        A = np.dot(g_val.T, H[:, :L])
        num = np.dot(A, A.T)
        denom = np.dot(np.dot(g_val.T, K_nq), g_val)

        return num / denom
